'''
Created on 30. mar. 2017

@author: mmpe
'''
import numpy as np
from wetb.signal.fit._spline_fit import spline_fit
from wetb.signal.filters._differentiation import differentiation


def fix_rotor_position(rotor_position, sample_frq, rotor_speed, fix_dt=None, plt=None):
    """Rotor position fitted with spline

    Parameters
    ----------
    rotor_position : array_like
        Rotor position [deg] (0-360)
    sample_frq : int or float
        Sample frequency [Hz]
    rotor_speed : array_like
        Rotor speed [RPM]
    fix_dt : int, float or None, optional
        Time distance [s] between spline fix points\n
        If None (default) a range of seconds is tested and the result that minimize the RMS 
        between differentiated rotor position fit and rotor speed is used.\n 
        Note that a significant speed up is achievable by specifying the parameter
    plt : PyPlot or None
        If PyPlot a visual interpretation is plotted

    Returns
    -------
    y : nd_array
        Fitted rotor position
    """

    from wetb.signal.subset_mean import revolution_trigger

    t = np.arange(len(rotor_position))
    indexes = revolution_trigger(rotor_position[:], sample_frq, rotor_speed, max_rev_diff=4)

    rp = rotor_position[:].copy()

    for i in indexes:
        rp[i:] += 360

    if fix_dt is None:
        fix_dt = find_fix_dt(rotor_position, sample_frq, rotor_speed)

    N = int(np.round(fix_dt * sample_frq))
    N2 = N // 2
    if plt:
        a = (rp.max() - rp.min()) / t.max()
        plt.plot(t / sample_frq, rp - t * a, label='Continuous rotor position (detrended)')

    points = []
    for j, i in enumerate(range(0, len(rp), N)):
        # indexes for subsets for overlapping a and b polynomials
        i1 = max(0, i - N2)
        i2 = min(len(rp), i + N2)

        # fit a polynomial
        if i1 < len(rp):
            poly_coef = np.polyfit(t[i1:i2] - t[i1], rp[i1:i2], 1)
            points.append((t[i], np.poly1d(poly_coef)(t[i] - t[i1])))
            if plt:
                plt.plot(t[i1:i2] / sample_frq, np.poly1d(poly_coef)(t[i1:i2] - t[i1]) - t[i1:i2]
                         * a, 'mc'[j % 2], label=('', "Line fit for points (detrended)")[j < 2])

    x, y = np.array(points).T

    if plt:
        plt.plot(x / sample_frq, y - x * a, '.', label='Fit points (detrended)')
        plt.plot(t / sample_frq, spline_fit(x, y)(t) - t * a, label='Spline (detrended)')
        plt.legend()
        plt.show()
    fit = spline_fit(x, y)(t)
    fit[differentiation(fit, "left") < 0] = np.nan
    return fit % 360


def find_fix_dt(rotor_position, sample_frq, rotor_speed, plt=None):
    """Find the optimal fix_dt parameter for fix_rotor_position (function above).
    Optimal is defined as the value that minimizes the sum of squared differences 
    between differentiated rotor position and rotor speed 

    Parameters
    ----------
    rotor_position : array_like
        Rotor position [deg] (0-360)
    sample_frq : int or float
        Sample frequency [Hz]
    rotor_speed : array_like
        Rotor speed [RPM]
    plt : pyplot or None
        If pyplot, a visual interpretation is plotted

    Returns
    -------
    y : int
        Optimal value for the fix_dt parameter for fix_rotor_position

    """
    from wetb.signal.filters import differentiation

    def err(i):
        drp = differentiation(fix_rotor_position(rotor_position, sample_frq, rotor_speed, i))
        rpm_pos = drp % 180 / 360 * sample_frq * 60
        return np.sum((rpm_pos - rotor_speed)**2)

    best = 27
    for step in [9, 3, 1]:

        x_lst = np.arange(-2, 3) * step + best
        res = [err(x) for x in x_lst]
        if plt is not None:
            plt.plot(x_lst, res, '.-')
        best = x_lst[np.argmin(res)]
    if plt is not None:
        plt.show()

    return best


def check_rotor_position(rotor_position, sample_frq, rotor_speed, max_rev_diff=1, plt=None):
    """Rotor position fitted with spline

    Parameters
    ----------
    rotor_position : array_like
        Rotor position [deg] (0-360)
    sample_frq : int or float
        Sample frequency [Hz]
    rotor_speed : array_like
        Rotor speed [RPM]
    fix_dt : int, float or None, optional
        Time distance [s] between spline fix points\n
        If None (default) a range of seconds is tested and the result that minimize the RMS 
        between differentiated rotor position fit and rotor speed is used.\n 
        Note that a significant speed up is achievable by specifying the parameter
    plt : PyPlot or None
        If PyPlot a visual interpretation is plotted

    Returns
    -------
    y : nd_array
        Fitted rotor position
    """

    from wetb.signal.subset_mean import revolution_trigger

    t = np.arange(len(rotor_position)) / sample_frq
    indexes = revolution_trigger(rotor_position[:], sample_frq, rotor_speed, max_rev_diff=max_rev_diff)

    rotor_position = rotor_position[:] % 360

    rp_fit = fix_rotor_position(rotor_position, sample_frq, rotor_speed, None)

    if plt:
        #a = (rp.max()-rp.min())/t.max()
        #plt.plot(t/sample_frq,rp-t*a,label='Continus rotor position (detrended)')
        f, axarr = plt.subplots(2)
        print(rp_fit)
        axarr[0].plot(t, rotor_position, label='rotor position')
        axarr[0].plot(indexes / sample_frq, rotor_position[indexes], '.', label='Revolution trigger')
        axarr[0].plot(t, rp_fit, label='Fitted rotor position')
        axarr[0].legend()
        axarr[1].plot(t, rotor_speed, label='Rotor speed')
        axarr[1].plot(indexes / sample_frq, 60 / (differentiation(indexes) /
                                                  sample_frq), label='Rotor speed (revolution average)')
        print(t[170:200])
        drp = differentiation(rp_fit)
        #drp[(drp<0)]= 0
        axarr[1].plot(t, drp % 180 / 360 * sample_frq * 60, label='Fix(calc from fitted rotor position)')
        axarr[1].legend()

    i1, i2 = indexes[0], indexes[-1]
    print("Rev from rotor position", np.sum(np.diff(rotor_position[i1:i2]) % 180) / 360)
    print("Rev from rotor speed", np.mean(rotor_speed[i1:i2]) * (i2 - i1) / 60 / sample_frq)
    print("Rev from fitted rotor position", np.sum(np.diff(rp_fit[i1:i2]) % 180) / 360)

    print("Mean RPM from rotor speed", np.mean(rotor_speed))
    print("Mean RPM from fitted rotor position", np.sum(
        np.diff(rp_fit[i1:i2]) % 180) / 360 / ((i2 - i1) / 60 / sample_frq))

    spr1 = (np.diff(indexes) / sample_frq)

    #rs1 = 60/( np.diff(indexes)/sample_frq)
    spr2 = np.array([60 / rotor_speed[i1:i2].mean() for i1, i2 in zip(indexes[:-1], indexes[1:])])

    err = spr1 - spr2
    print(err.max())



'''
Created on 05/11/2015

@author: MMPE
'''
import unittest
import os

from numpy import testing

from wetb.hawc2.st_file import StFile


testfilepath = os.path.join(os.path.dirname(__file__), 'test_files/')  # test file path
class TestStFile(unittest.TestCase):


    def test_stfile(self):
        st = StFile(testfilepath + 'DTU_10MW_RWT_Blade_st.dat')
        self.assertEqual(st.radius_st()[2], 3.74238)
        self.assertEqual(st.radius_st(3), 3.74238)
        self.assertEqual(st.x_e(67.7351), 4.4320990737400E-01)
        self.assertEqual(st.E(3.74238, 1, 1), 1.2511695058500E+10)
        self.assertEqual(st.E(3.74238, 1, 2), 1.2511695058500E+27)

    def test_fpmfile(self):
        st = StFile(os.path.join(testfilepath, 'IEA_15MW_RWT_Blade_fpm_st.dat'))
        self.assertEqual(st.radius_st()[2], 5.857445433495505)
        self.assertEqual(st.radius_st(10), 11.71489728616879)
        self.assertEqual(st.x_e(67.7351), 0.7150636117520498)
        self.assertEqual(st.K_11(3.5, 1, 1), 5572685723.993099)
        self.assertEqual(st.K_11(3.5, 1, 2), 5572685723.993099)

    def test_stfile_interpolate(self):
        st = StFile(testfilepath + 'DTU_10MW_RWT_Blade_st.dat')
        self.assertAlmostEqual(st.x_e(72.2261), 0.381148048)
        self.assertAlmostEqual(st.y_e(72.2261), 0.016692967)

    def test_save(self):
        fname = os.path.join(testfilepath, 'DTU_10MW_RWT_Blade_st.dat')
        fname2 = os.path.join(testfilepath, 'DTU_10MW_RWT_Blade_st2.dat')
        st = StFile(fname)
        st.save(fname2, encoding='utf-8', precision='%20.12e')
        st2 = StFile(fname2)
        self.assertEqual(len(st.main_data_sets), len(st2.main_data_sets))
        self.assertEqual(len(st.main_data_sets[1]), len(st2.main_data_sets[1]))
        for k in st.main_data_sets[1]:
            testing.assert_almost_equal(st.main_data_sets[1][k],
                                        st2.main_data_sets[1][k], decimal=12)
        os.remove(fname2)

    def test_save_fpm(self):
        fname = os.path.join(testfilepath, 'IEA_15MW_RWT_Blade_fpm_st.dat')
        fname2 = os.path.join(testfilepath, 'IEA_15MW_RWT_Blade_fpm_st2.dat')
        st = StFile(fname)
        st.save(fname2, encoding='utf-8', precision='% 24.15e')
        st2 = StFile(fname2)
        self.assertEqual(len(st.main_data_sets), len(st2.main_data_sets))
        self.assertEqual(len(st.main_data_sets[1]), len(st2.main_data_sets[1]))
        for k in st.main_data_sets[1]:
            testing.assert_almost_equal(st.main_data_sets[1][k],
                                        st2.main_data_sets[1][k], decimal=12)
        os.remove(fname2)


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()




import numpy as np
from scipy.interpolate import interp1d


def bin_fit(x, y, bins=10, kind='linear', bin_func=np.nanmean, bin_min_count=3, lower_upper='discard'):
    """Fit observations based on bin statistics

    Parameters
    ---------
    x : array_like
        x observations
    y : array_like
        y observations
    bins : int, array_like or (int, int)
        if int: <bins> binx evenly distributed on the x-axis
        if (xbins,ybins): <xbins> and <ybins> evenly distributed on the x and y axis respectively\n
        Note that ybins only make sense if every y-value maps to a single x-value
    kind : int or string
        degree of polynomial for fit (argument passed to scipy.interpolate.interpolate.interp1d)
    bin_func : function, optional
        Statistic function to apply on bins, default is nanmean
    bin_min_count : int, optional
        Minimum number of observations in bins to include
        Default is 3
    lower_upper : str, int, (str,str), (int,int)
        How to handle observations below and above first and last bin values. Can be:\n
        - "discard":
        - "extrapolate":
        - int: Set f(max(x)) to mean of first/last int observations


    Returns
    -------
    bin_x, fit_function

    """
    x, y = np.array(x[:]), np.array(y[:])
    if isinstance(bins, int):
        bins = np.linspace(np.nanmin(x), np.nanmax(x) + 1e-10, bins + 1)
    elif isinstance(bins, tuple) and len(bins) == 2 and isinstance(bins[0], int) and isinstance(bins[1], int):
        xbins, ybins = bins
        if xbins > 0:
            xbinsx = np.linspace(np.nanmin(x), np.nanmax(x) + 1e-10, xbins + 1)
        else:
            xbinsx = []
        if ybins > 0:
            x1, f1 = bin_fit(y, x, kind=1, bins=ybins)
            xbinsy = f1(x1)
        else:
            xbinsy = []
        #x2, f2 = bin_fit(x,y, kind=1, bins=xbins)
        bins = sorted(np.r_[xbinsx, xbinsy])

    digitized = np.digitize(x, bins)
    digitized[np.isnan(x) | np.isnan(y)] = -1

    masks = [digitized == i for i in range(1, len(bins))]
    import warnings
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        bin_x = np.array([np.nanmean(x[mask]) for mask in masks])
        bin_y = np.array([bin_func(y[mask]) for mask in masks])
        bin_count = np.array([np.sum(mask) for mask in masks])
    #bin_x_fit, bin_y = [b[bin_count >= bin_min_count] for b in [bin_x, bin_y]]
    bin_x_fit = bin_x
    m = np.isnan(bin_x_fit)
    bin_x_fit[m] = ((bins[:-1] + bins[1:]) / 2)[m]
    bin_y_fit = bin_y.copy()
    bin_y_fit[bin_count < bin_min_count] = np.nan

    if isinstance(lower_upper, (str, int)):
        lower = upper = lower_upper
    else:
        lower, upper = lower_upper

    # Add value to min(x)
    if bin_x_fit[0] > np.nanmin(x) or np.isnan(bin_y_fit[0]):
        if lower == 'extrapolate':

            bin_y_fit = np.r_[bin_y_fit[0] - (bin_x_fit[0] - np.nanmin(x)) *
                              (bin_y_fit[1] - bin_y_fit[0]) / (bin_x_fit[1] - bin_x_fit[0]), bin_y_fit]
            bin_x_fit = np.r_[np.nanmin(x), bin_x_fit]
        elif lower == "discard":
            pass
        elif isinstance(lower, int):
            bin_y_fit = np.r_[np.mean(y[~np.isnan(x)][np.argsort(x[~np.isnan(x)])[:lower]]), bin_y_fit]
            bin_x_fit = np.r_[np.nanmin(x), bin_x_fit]
        else:
            raise NotImplementedError("Argument for handling lower observations, %s, not implemented" % lower)

    # add value to max(x)
    if bin_x_fit[-1] < np.nanmax(x) or np.isnan(bin_y_fit[-1]):
        if upper == 'extrapolate':
            bin_y_fit = np.r_[bin_y_fit, bin_y_fit[-1] + (np.nanmax(x) - bin_x_fit[-1])
                              * (bin_y_fit[-1] - bin_y_fit[-2]) / (bin_x_fit[-1] - bin_x_fit[-2])]
            bin_x_fit = np.r_[bin_x_fit, np.nanmax(x)]
        elif upper == "discard":
            pass
        elif isinstance(upper, int):
            bin_y_fit = np.r_[bin_y_fit, np.mean(y[~np.isnan(x)][np.argsort(x[~np.isnan(x)])[-upper:]])]
            bin_x_fit = np.r_[bin_x_fit, np.nanmax(x)]
        else:
            raise NotImplementedError("Argument for handling upper observations, %s, not implemented" % upper)

    return bin_x_fit, _interpolate_fit(bin_x_fit, bin_y_fit, kind)


def perpendicular_bin_fit(x, y, bins=30, fit_func=None, bin_min_count=3, plt=None):
    """Fit a curve to the values, (x,y) using bins that are perpendicular to an initial fit

    Parameters
    ---------
    x : array_like
        x observations
    y : array_like
        y observations
    bins : int
        Number of perpendicular bins
    fit_func : function(x,y) -> (x,y) or None
        Initial fit function
        If None, bin_fit with same number of bins are used
    bin_min_count : int, optional
        Minimum number of observations in bins to include
        Default is 3

    plt : pyplot or None
        If pyplot the fitting process is plotted on plt

    Returns
    -------
    fit_x, fit_y
    """

    if fit_func is None:
        def fit_func(x, y): return bin_fit(x, y, bins, bin_func=np.nanmean)

    x, y = [v[~np.isnan(x) & ~np.isnan(y)] for v in [x, y]]

    bfx, f = fit_func(x, y)
    bfy = f(bfx)

    bfx, bfy = [v[~np.isnan(bfx) & ~np.isnan(bfy)] for v in [bfx, bfy]]
    if plt:
        x_range, y_range = [v.max() - v.min() for v in [x, y]]
        plt.ylim([y.min() - y_range * .1, y.max() + y_range * .1])
        plt.xlim([x.min() - x_range * .1, x.max() + x_range * .1])

    # divide curve into N segments of same normalized curve length
    xg, xo = np.nanmax(bfx) - np.nanmin(bfx), np.nanmin(bfx)
    yg, yo = np.nanmax(bfy) - np.nanmin(bfy), np.nanmin(bfy)
    nbfx = (bfx - xo) / xg
    nbfy = (bfy - yo) / yg
    l = np.cumsum(np.sqrt(np.diff(nbfx)**2 + np.diff(nbfy)**2))
    nx, ny = [np.interp(np.linspace(l[0], l[-1], bins + 1), l, (xy[1:] + xy[:-1]) / 2) for xy in [nbfx, nbfy]]

    last = (-1, 0)

    pc = []
    used = np.zeros_like(x).astype(bool)
    for i in range(0, len(nx)):
        i1, i2 = max(0, i - 1), min(len(nx) - 1, i + 1)
        a = -(nx[i2] - nx[i1]) / (ny[i2] - ny[i1])
        b = (ny[i] - (a * nx[i])) * yg + yo
        a *= yg / xg
        x_ = [np.nanmin(x), np.nanmax(x)]
        m1 = np.sign(last[0]) * y < np.sign(last[0]) * ((x - xo) * last[0] + last[1])
        m2 = np.sign(a) * y > np.sign(a) * (a * (x - xo) + b)
        m = m1 & m2 & ~used
        if plt:
            plt.plot(x_, ((a) * (x_ - xo)) + b)
            plt.plot(x[m], y[m], '.')

        if np.sum(m) >= bin_min_count:
            pc.append((np.median(x[m]), np.median(y[m])))
            used = used | m
        last = (a, b)
    #bfx,bfy = zip(*pc)
    if plt:
        pbfx, pbfy = np.array(pc).T
        plt.plot(bfx, bfy, 'orange', label='initial_fit')
        plt.plot(pbfx, pbfy, 'gray', label="perpendicular fit")
        plt.legend()
    #PlotData(None, bfx,bfy)
    bin_x_fit, bin_y_fit = np.array(pc).T
    return bin_x_fit, _interpolate_fit(bin_x_fit, bin_y_fit, kind="linear")

# Create mean function


def _interpolate_fit(bin_x_fit, bin_y_fit, kind='linear'):
    def fit(x):
        x = np.atleast_1d(x)[:].copy().astype(float)
        x[x < bin_x_fit[0]] = np.nan
        x[x > bin_x_fit[-1]] = np.nan
        m = ~(np.isnan(bin_x_fit) | np.isnan(bin_y_fit))
        return interp1d(bin_x_fit[m], bin_y_fit[m], kind)(x[:])
    return fit



import unittest
import os
#import shutil

import numpy as np
import pandas as pd

from wetb.prepost.GenerateDLCs import GenerateDLCCases


class Template(unittest.TestCase):
    def setUp(self):
        self.basepath = os.path.dirname(__file__)


class TestGenerateDLCCases(Template):

    def test_dlcs(self):
        # manually configure paths, HAWC2 model root path is then constructed as
        # p_root_remote/PROJECT/sim_id, and p_root_local/PROJECT/sim_id
        # adopt accordingly when you have configured your directories differently
        p_root = os.path.join(self.basepath, 'data/demo_gendlc/')

        # project name, sim_id, master file name
        dlc_master = os.path.join(p_root, 'DLCs.xlsx')
        dlc_folder = os.path.join(p_root, 'DLCs')
        dlc_gen12 = os.path.join(dlc_folder, 'DLC12.xlsx')
        dlc_gen13 = os.path.join(dlc_folder, 'DLC13.xlsx')

        DLB = GenerateDLCCases()
        DLB.execute(filename=dlc_master, folder=dlc_folder)

        df12 = pd.read_excel(dlc_gen12)
        # df12.to_csv('data/demo_gendlc/ref/DLC12.csv', index=False)
        df12_ref = pd.read_csv(os.path.join(p_root, 'ref/DLC12.csv'))
        # df12_ref2 = pd.read_excel(p2)[df12.columns]
        pd.testing.assert_frame_equal(df12, df12_ref)

        # df2 = df[['[Case id.]', '[wdir]', '[wsp]', '[seed]', '[wave_seed]']]
        self.assertEqual(df12['[ReferenceWindSpeed]'].unique(), np.array([44]))
        self.assertEqual(df12['[t0]'].unique(), np.array([100]))
        self.assertEqual(len(df12['[Case id.]'].unique()), 2*3*3*2)
        self.assertEqual(df12['[Case id.]'].values[0],
                         'DLC12_wsp04_wdir010_sa1001_sw0101')

        df13 = pd.read_excel(dlc_gen13)
        # df13.to_csv('data/demo_gendlc/ref/DLC13.csv', index=False)
        df13_ref = pd.read_csv(os.path.join(p_root, 'ref/DLC13.csv'))
        pd.testing.assert_frame_equal(df13, df13_ref)


if __name__ == "__main__":
    unittest.main()



import os
import numpy as np
import struct
from itertools import takewhile


def load_output(filename):
    """Load a FAST binary or ascii output file

    Parameters
    ----------
    filename : str
        filename

    Returns
    -------
    data : ndarray
        data values
    info : dict
        info containing:
            - name: filename
            - description: description of dataset
            - attribute_names: list of attribute names
            - attribute_units: list of attribute units
    """

    assert os.path.isfile(filename), "File, %s, does not exists" % filename
    with open(filename, 'r') as f:
        try:
            f.readline()
        except UnicodeDecodeError:
            return load_binary_output(filename)
    return load_ascii_output(filename)


def load_ascii_output(filename):
    with open(filename) as f:
        info = {}
        info['name'] = os.path.splitext(os.path.basename(filename))[0]
        # Header is whatever is before the keyword `time`
        in_header = True
        header = []
        while in_header:
            l = f.readline()
            if not l:
                raise Exception('Error finding the end of FAST out file header. Keyword Time missing.')
            in_header = (l + ' dummy').lower().split()[0] != 'time'
            if in_header:
                header.append(l)
            else:
                info['description'] = header
                info['attribute_names'] = l.split()
                info['attribute_units'] = [unit[1:-1] for unit in f.readline().split()]

        # Data, up to end of file or empty line (potential comment line at the end)
        data = np.array([l.strip().split() for l in takewhile(
            lambda x: len(x.strip()) > 0, f.readlines())]).astype(float)
        return data, info


def load_binary_output(filename, use_buffer=True):
    """
    03/09/15: Ported from ReadFASTbinary.m by Mads M Pedersen, DTU Wind
    24/10/18: Low memory/buffered version by E. Branlard, NREL

    Info about ReadFASTbinary.m:
    % Author: Bonnie Jonkman, National Renewable Energy Laboratory
    % (c) 2012, National Renewable Energy Laboratory
    %
    %  Edited for FAST v7.02.00b-bjj  22-Oct-2012
    """
    def fread(fid, n, type):
        fmt, nbytes = {
            'uint8': ('B', 1),
            'int16': ('h', 2),
            'int32': ('i', 4),
            'float32': ('f', 4),
            'float64': ('d', 8)}[type]
        return struct.unpack(fmt * n, fid.read(nbytes * n))

    def freadRowOrderTableBuffered(fid, n, type_in, nCols, nOff=0, type_out='float64'):
        """
        Reads of row-ordered table from a binary file.

        Read `n` data of type `type_in`, assumed to be a row ordered table of `nCols` columns.
        Memory usage is optimized by allocating the data only once.
        Buffered reading is done for improved performances (in particular for 32bit python)

        `nOff` allows for additional column space at the begining of the storage table.
        Typically, `nOff=1`, provides a column at the beginning to store the time vector.

        @author E.Branlard, NREL

        """
        fmt, nbytes = {
            'uint8': (
                'B', 1), 'int16': (
                'h', 2), 'int32': (
                'i', 4), 'float32': (
                    'f', 4), 'float64': (
                        'd', 8)}[type_in]
        nLines = int(n / nCols)
        GoodBufferSize = 4096 * 40
        nLinesPerBuffer = int(GoodBufferSize / nCols)
        BufferSize = nCols * nLinesPerBuffer
        nBuffer = int(n / BufferSize)
        # Allocation of data
        data = np.zeros((nLines, nCols + nOff), dtype=type_out)
        # Reading
        try:
            nIntRead = 0
            nLinesRead = 0
            while nIntRead < n:
                nIntToRead = min(n - nIntRead, BufferSize)
                nLinesToRead = int(nIntToRead / nCols)
                Buffer = np.array(struct.unpack(fmt * nIntToRead, fid.read(nbytes * nIntToRead)))
                Buffer = Buffer.reshape(-1, nCols)
                data[nLinesRead:(nLinesRead + nLinesToRead), nOff:(nOff + nCols)] = Buffer
                nLinesRead = nLinesRead + nLinesToRead
                nIntRead = nIntRead + nIntToRead
        except Exception:
            raise Exception('Read only %d of %d values in file: %s' % (nIntRead, n, filename))
        return data

    FileFmtID_WithTime = 1  # % File identifiers used in FAST
    FileFmtID_WithoutTime = 2
    FileFmtID_NoCompressWithoutTime = 3
    FileFmtID_ChanLen_In = 4  # time channel and channel length is not included

    with open(filename, 'rb') as fid:
        FileID = fread(fid, 1, 'int16')[0]  # ;             % FAST output file format, INT(2)
        if FileID not in [FileFmtID_WithTime, FileFmtID_WithoutTime,
                          FileFmtID_ChanLen_In, FileFmtID_NoCompressWithoutTime]:
            raise Exception('FileID not supported {}. Is it a FAST binary file?'.format(FileID))

        if FileID == FileFmtID_ChanLen_In:
            LenName = fread(fid, 1, 'int16')[0]  # Number of characters in channel names and units
        else:
            LenName = 10  # default number of characters per channel name

        NumOutChans = fread(fid, 1, 'int32')[0]  # ;             % The number of output channels, INT(4)
        NT = fread(fid, 1, 'int32')[0]  # ;             % The number of time steps, INT(4)

        if FileID == FileFmtID_WithTime:
            TimeScl = fread(fid, 1, 'float64')  # ;           % The time slopes for scaling, REAL(8)
            TimeOff = fread(fid, 1, 'float64')  # ;           % The time offsets for scaling, REAL(8)
        else:
            TimeOut1 = fread(fid, 1, 'float64')  # ;           % The first time in the time series, REAL(8)
            TimeIncr = fread(fid, 1, 'float64')  # ;           % The time increment, REAL(8)

        if FileID == FileFmtID_NoCompressWithoutTime:
            ColScl = np.ones(NumOutChans)
            ColOff = np.zeros(NumOutChans)
        else:
            ColScl = fread(fid, NumOutChans, 'float32')  # ; % The channel slopes for scaling, REAL(4)
            ColOff = fread(fid, NumOutChans, 'float32')  # ; % The channel offsets for scaling, REAL(4)

        LenDesc = fread(fid, 1, 'int32')[0]  # ;  % The number of characters in the description string, INT(4)
        DescStrASCII = fread(fid, LenDesc, 'uint8')  # ;  % DescStr converted to ASCII
        DescStr = "".join(map(chr, DescStrASCII)).strip()

        ChanName = []  # initialize the ChanName cell array
        for iChan in range(NumOutChans + 1):
            ChanNameASCII = fread(fid, LenName, 'uint8')  # ; % ChanName converted to numeric ASCII
            ChanName.append("".join(map(chr, ChanNameASCII)).strip())

        ChanUnit = []  # initialize the ChanUnit cell array
        for iChan in range(NumOutChans + 1):
            ChanUnitASCII = fread(fid, LenName, 'uint8')  # ; % ChanUnit converted to numeric ASCII
            ChanUnit.append("".join(map(chr, ChanUnitASCII)).strip()[1:-1])

        #    %-------------------------
        #    % get the channel time series
        #    %-------------------------

        nPts = NT * NumOutChans  # ;           % number of data points in the file

        if FileID == FileFmtID_WithTime:
            PackedTime = fread(fid, NT, 'int32')  # ; % read the time data
            cnt = len(PackedTime)
            if cnt < NT:
                raise Exception('Could not read entire %s file: read %d of %d time values' % (filename, cnt, NT))

        if use_buffer:
            # Reading data using buffers, and allowing an offset for time column (nOff=1)
            if FileID == FileFmtID_NoCompressWithoutTime:
                data = freadRowOrderTableBuffered(fid, nPts, 'float64', NumOutChans, nOff=1, type_out='float64')
            else:
                data = freadRowOrderTableBuffered(fid, nPts, 'int16', NumOutChans, nOff=1, type_out='float64')
        else:
            # NOTE: unpacking huge data not possible on 32bit machines
            if FileID == FileFmtID_NoCompressWithoutTime:
                PackedData = fread(fid, nPts, 'float64')  # ; % read the channel data
            else:
                PackedData = fread(fid, nPts, 'int16')  # ; % read the channel data
            cnt = len(PackedData)
            if cnt < nPts:
                raise Exception('Could not read entire %s file: read %d of %d values' % (filename, cnt, nPts))
            data = np.array(PackedData).reshape(NT, NumOutChans)
            del PackedData

    if FileID == FileFmtID_WithTime:
        time = (np.array(PackedTime) - TimeOff) / TimeScl
    else:
        time = TimeOut1 + TimeIncr * np.arange(NT)

    #    %-------------------------
    #    % Scale the packed binary to real data
    #    %-------------------------
    if use_buffer:
        # Scaling Data
        for iCol in range(NumOutChans):
            data[:, iCol + 1] = (data[:, iCol + 1] - ColOff[iCol]) / ColScl[iCol]
        # Adding time column
        data[:, 0] = time
    else:
        # NOTE: memory expensive due to time conversion, and concatenation
        data = (data - ColOff) / ColScl
        data = np.concatenate([time.reshape(NT, 1), data], 1)

    info = {'name': os.path.splitext(os.path.basename(filename))[0],
            'description': DescStr,
            'attribute_names': ChanName,
            'attribute_units': ChanUnit}
    return data, info



test = "TEST"
__version__ = '0.0.10'
# try:
#     import pkg_resources
#     __version__ = pkg_resources.safe_version(pkg_resources.get_distribution(__name__).version)
# except:
#     __version__ = 'unknown'



'''
Created on 22. jun. 2017

@author: mmpe
'''
import os

import numpy as np
from wetb.hawc2.htc_file import HTCFile
from wetb.wind.turbulence import mann_turbulence


class TurbulenceFile(object):
    def __init__(self, filename, Nxyz, dxyz, transport_speed=10, mean_wsp=0, center_position=(0, 0, -20)):
        self.filename = filename
        self.Nxyz = Nxyz
        self.dxyz = dxyz
        self.transport_speed = transport_speed
        self.mean_wsp = mean_wsp
        self.center_position = center_position
        self.data = mann_turbulence.load(filename, Nxyz)

    @property
    def data3d(self):
        return self.data.reshape(self.Nxyz)

    @staticmethod
    def load_from_htc(htcfilename, modelpath=None, type='mann'):
        htc = HTCFile(htcfilename, modelpath)

        Nxyz = np.array([htc.wind[type]['box_dim_%s' % uvw][0] for uvw in 'uvw'])
        dxyz = np.array([htc.wind[type]['box_dim_%s' % uvw][1] for uvw in 'uvw'])
        center_position = htc.wind.center_pos0.values
        wsp = htc.wind.wsp
        return [TurbulenceFile(os.path.join(htc.modelpath, htc.wind[type]['filename_%s' % uvw][0]), Nxyz, dxyz, wsp, (0, wsp)[uvw == 'u'], center_position) for uvw in 'uvw']


if __name__ == '__main__':
    pass




import os


def repl(path):
    return path.replace("\\", "/")


def abspath(path):
    return repl(os.path.abspath(path))


def relpath(path, start=None):
    return repl(os.path.relpath(path, start))


def realpath(path):
    return repl(os.path.realpath(path))


def pjoin(*path):
    return repl(os.path.join(*path))


def fixcase(path):
    path = realpath(str(path)).replace("\\", "/")
    p, rest = os.path.splitdrive(path)
    p += "/"
    for f in rest[1:].split("/"):
        f_lst = [f_ for f_ in os.listdir(p) if f_.lower() == f.lower()]
        if len(f_lst) > 1:
            # use the case sensitive match
            f_lst = [f_ for f_ in f_lst if f_ == f]
        if len(f_lst) == 0:
            raise IOError("'%s' not found in '%s'" % (f, p))
        # Use matched folder
        p = pjoin(p, f_lst[0])
    return p


def normpath(path):
    return repl(os.path.normpath(path))


def cluster_path(path):
    try:
        import win32wnet
        drive, folder = os.path.splitdrive(abspath(path))
        path = abspath(win32wnet.WNetGetUniversalName(drive, 1) + folder)
    except Exception:
        path = repl(path)
    path = path.replace("//jess.dtu.dk", "/home")
    path = path.replace("//mimer.risoe.dk/aiolos", "/mnt/aiolos")
    path = path.replace("//mimer.risoe.dk", "/mnt/mimer")
    return path



import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from wetb.utils.rotation import axis2axis_angle, axis_angle2quaternion, quaternion2matrix, s2matrix, matrix2quaternion,\
    quaternion2axis_angle, axis_angle2axis, matrix2axis


def init_plot(ref_coo='gl', rot=(180, 180)):
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    plot_coo(ref_coo, np.eye(3) * 1.5, lw=1, alpha=1)
    for xyz in 'xyz':
        getattr(ax, 'set_%slim' % xyz)([-2, 2])
        getattr(ax, 'set_%slabel' % xyz)(xyz)

    ax.view_init(rot[0] + 10, rot[1] - 10)


def plot_coo(coo, A, origo=(0, 0, 0), lw=1, alpha=1):
    ax = plt.gca()
    for v, c, xyz in zip(np.asarray(A).T, 'rgb', 'xyz'):
        ax.quiver(*origo, *v, color=c, lw=lw, alpha=alpha, arrow_length_ratio=.2, zorder=-32)
        ax.text(*(np.array(origo) + v * 1.1), "$%s_{%s}$" % (xyz, coo),
                alpha=alpha, fontsize=12, fontweight='bold', zorder=32)


def plot_axis_rotation(rot_coo, axis, origo=(0, 0, 0), ref_coo=np.eye(3), deg=False):

    axis_angle = axis2axis_angle(axis)
    mat = quaternion2matrix(axis_angle2quaternion(axis_angle, deg))
    mat = np.dot(ref_coo, mat)
    axis = np.dot(ref_coo, axis_angle[:3])
    xyz = np.array([origo, np.array(origo) + axis * 1.5])
    plt.gca().plot(*xyz.T, '-.k', lw=1)
    plot_coo(rot_coo, mat, origo)
    return mat


def plot_matrix_rotation(rot_name, matrix, origo, ref_coo=np.eye(3)):
    mat = np.dot(ref_coo, matrix)
    plot_coo(rot_name, mat, origo)


def plot_body(nodes):
    ax = plt.gca()
    ax.plot(*np.array(nodes).T, alpha=0.2, lw=10)
    ax.plot(*np.array(nodes).T, '.k', alpha=1)


def set_aspect_equal():
    ax = plt.gca()
    xyz_lim = [getattr(ax, 'get_%slim' % xyz)() for xyz in 'xyz']
    max_range = np.max([np.abs(np.subtract(*lim)) for lim in xyz_lim]) / 2
    mid = np.mean(xyz_lim, 1)
    for xyz, m in zip('xyz', mid):
        getattr(ax, "set_%slim" % xyz)([m - max_range, m + max_range])
    plt.tight_layout()


def vs_array(s, shape):
    arr = np.array(s.strip().replace("D", "E").split("\n"), dtype=float)
    if len(np.atleast_1d(shape)) == 2:
        arr = arr.reshape(shape[::-1]).T
    return arr


if __name__ == '__main__':
    TBD = vs_array("""-0.991775138730704
-1.644972253859245D-002
0.126931008126843
1.644972253859245D-002
0.967100554922815
0.253862016253685
-0.126931008126843
0.253862016253685
-0.958875693653519""", (3, 3))
    TUB = vs_array("""-1.00000000000000
0.000000000000000D+000
0.000000000000000D+000
0.000000000000000D+000
1.00000000000000
0.000000000000000D+000
0.000000000000000D+000
0.000000000000000D+000
-1.00000000000000""", (3, 3))
    TDU = vs_array("""0.991775138730704
-1.644972253859245D-002
0.126931008126843
-1.644972253859245D-002
0.967100554922815
0.253862016253685
-0.126931008126843
-0.253862016253685
0.958875693653519""", (3, 3))
    print(np.dot(TBD.T, TUB.T) - TDU)
    print(TDU)
    rot_bd = matrix2axis(TBD)
    rot_bu = matrix2axis(TUB)
    rot_du = matrix2axis(TDU)

    print(np.dot(TUB, matrix2axis(np.dot(TUB.T, TBD.T))))
    print(matrix2axis(TDU))

    print(np.dot(TUB.T, TBD.T).T)



'''
Created on 10/03/2014

@author: MMPE
'''
import glob
import re

import os


DEBUG = False


def pexec(args, cwd=None, shell=False):
    """
    usage: errorcode, stdout, stderr, cmd = pexec("MyProgram.exe arg1, arg2", r"c:\tmp\")

    """
    proc, cmd = process(args, cwd)
    errorcode, stdout, stderr = exec_process(proc)
    return errorcode, stdout, stderr, cmd


def process(args, cwd=None, shell=False):
    import subprocess
    if not isinstance(args, (list, tuple)):
        args = [args]
    args = [str(arg) for arg in args]
    for i in range(len(args)):
        if os.path.exists(args[i]):
            args[i] = str(args[i]).replace('/', os.path.sep).replace('\\', os.path.sep).replace('"', '')

    cmd_args = subprocess.list2cmdline(args)
    if cwd and os.path.isfile(cwd):
        cwd = os.path.dirname(cwd)
    if cwd:
        cmd_cwd = "cd %s && " % cwd
    else:
        cmd_cwd = ""

    if os.name == 'nt':
        cmd = '%s /c "%s%s"' % (os.environ.get("COMSPEC", "cmd.exe"), cmd_cwd, cmd_args)
    else:
        cmd = '%s%s' % (cmd_cwd, cmd_args)

    return subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=shell, cwd=cwd), cmd


def exec_process(process):
    stdout, stderr = process.communicate()
    errorcode = process.returncode

    return errorcode, stdout.decode(), stderr.decode()



# -*- coding: utf-8 -*-
"""
Created on Sun Mar 25 16:15:25 2018

@author: shfe@dtu.dk

The model proposed by Leblanc
"""
import matplotlib.pyplot as plt
import numpy as np
plt.close('all')

class LeblancModel(object):
    
    
    def __init__(self, xi_b, xi_c, N):
        """
        Initialize the input parameters
        """
        
        self.xi_b = xi_b
        self.xi_c = xi_c
        self.N = N
        self.alpha = 0.31
        
    def coeff(self, rd = 0.38):
        """
        Calculate the parameter alpha
        """
        # Calculate the Tb, Tc
        if not 0 <= self.xi_b <= 1:
            raise ValueError('The maximum load ratio should between 0 to 1')
        
        if not -1 <= self.xi_c <= 1:
            raise ValueError('The cyclic load ratio should between -1 to 1')
        
        if rd == 0.38:
            if self.xi_b <= 0.051:
                Tb = 0
            else:
                Tb = 0.4238*self.xi_b - 0.0217
        
        elif rd == 0.04:
            if self.xi_b <= 0.1461:
                Tb = 0
            else:
                Tb = 0.3087*self.xi_b - 0.0451
                
        else:
            raise ValueError('The relative density value is not validated yet')
                
                
        if -1 <= self.xi_c < -0.65:
            Tc = 13.71*self.xi_c + 13.71
            
        elif -0.65 <= self.xi_c < 0:
            Tc = -5.54*self.xi_c + 1.2
            
        else:
            Tc = -1.2*self.xi_c + 1.2
        
        return Tb, Tc 
    
    def evoluation(self, y1):
        """
        Evoluation model for cyclic accumulated rotation
        """
        
        Tb, Tc = self.coeff()
        yn = y1 * Tb * Tc * self.N**0.31
        
        return yn
        
    def equivalentnum(self, y1, yn):
        """
        Equivalent number for cylces
        """
        Tb, Tc = self.coeff()
        num = (yn/(y1*Tb*Tc))**(1/self.alpha)
        
        return num

def superposition(xi_b, xi_c, N_cases, y1_cases, s0 = 0, y0 = 0):

    
    # Check the array equal number of items
    if not (np.size(xi_b) == np.size(xi_c) and np.size(xi_b) == np.size(N_cases)):
        raise ValueError('Size for input should be identical')
#    print(s0)
    dyn = np.zeros(np.size(xi_b))
    yn = np.zeros(np.size(xi_b))
    y_static = np.zeros(np.size(xi_b))
    for i, (b, c, N, y1) in enumerate(zip(xi_b, xi_c, N_cases, y1_cases)):
        
        if i==0:
            
            em = LeblancModel(b, c, N)
            Tb, Tc = em.coeff()
            if s0 == 0:
                dyn[i] = em.evoluation(y1)
            else:
                N_eq = (s0/(y1*Tb*Tc))**(1/em.alpha)
                if N_eq < 1e8:
                    dyn[i] = (y1*Tb*Tc)*(N+N_eq)**em.alpha
                else:
                    dyn[i] = s0
            
            y_static[i] = max(y1, y0)
                               
        else:
            em = LeblancModel(b, c, N)
            Tb, Tc = em.coeff()
            N_eq = (dyn[i-1]/(y1*Tb*Tc))**(1/em.alpha)
            if N_eq < 1e8:
                dyn[i] = (y1*Tb*Tc)*(N_eq + N)**em.alpha
            else:
                dyn[i] = dyn[i-1]
                
            y_static[i] = max(y_static[i-1], y1)
        
        yn[i] = dyn[i] + y_static[i]
        
    return dyn, y_static, yn
    
def coeff_super(xi_b, xi_c, N_cases):
    # Check the alpha value
    if not (np.size(xi_b) == np.size(xi_c) and np.size(xi_b) == np.size(N_cases)):
        raise ValueError('Size for input should be identical')
    
    Tb_list = np.zeros(np.size(xi_b))
    Tc_list = np.zeros(np.size(xi_c))
    for i, (b, c, N) in enumerate(zip(xi_b, xi_c, N_cases)):
        
        em = LeblancModel(b, c, N)
        Tb, Tc = em.coeff()
        Tb_list[i] = Tb
        Tc_list[i] = Tc
        
    return Tb_list, Tc_list
    
    
def moment_ult(L, D, gamma):
    """
    Ultimate static moment
    
    Parameters
    ----------------
    L -- Pile length
    D -- Pile diameter
    gamma -- Submerged unit weight
    
    Output
    ----------------
    m_ult -- Ultimate static moment
    """
    
    
    m_ult = L**3*D*gamma*0.6
    
    return m_ult
    
def rot_moment(m, L, D, gamma, pa = 101, rd  = 0.04):
    """
    Ultimate static moment
    
    Parameters
    ----------------
    L -- Pile length
    D -- Pile diameter
    gamma -- Submerged unit weight
    m -- Moment
    
    Output
    ----------------
    theta - Rotational angle in radians
    """
    
    if rd == 0.04:
        m_ult = L**3*D*gamma*0.6/1e3   # unit: MN
        theta_norm = 0.038 * (m/m_ult)**2.33
        theta = theta_norm*np.sqrt((L*gamma)/pa)
    
    else:
        m_ult = L**3*D*gamma*1.2/1e3    # unit: MN
        theta_norm = 0.042 * (m/m_ult)**1.92
        theta = theta_norm*np.sqrt((L*gamma)/pa)

    return theta




'''
Created on 24/04/2014

@author: MMPE
'''
import types
import os
import numpy as np


stc = "r m x_cg y_cg ri_x ri_y x_sh y_sh E G I_x I_y I_p k_x k_y A pitch x_e y_e"
fpm = 'r m x_cg y_cg ri_x ri_y pitch x_e y_e K_11 K_12 K_13 K_14 K_15 K_16 K_22'
fpm += ' K_23 K_24 K_25 K_26 K_33 K_34 K_35 K_36 K_44 K_45 K_46 K_55 K_56 K_66'

class StFile(object):
    """Read HAWC2 St (beam element structural data) file

    Methods are autogenerated for:

    - r : curved length distance from main_body node 1 [m]
    - m : mass per unit length [kg/m]
    - x_cg : xc2-coordinate from C1/2 to mass center [m]
    - y_cg : yc2-coordinate from C1/2 to mass center [m]
    - ri_x : radius of gyration related to elastic center. Corresponds to rotation about principal bending xe axis [m]
    - ri_y : radius of gyration related to elastic center. Corresponds to rotation about principal bending ye axis [m]
    - xs : xc2-coordinate from C1/2 to shear center [m]. The shear center is the point where external forces only contributes to pure bending and no torsion.
    - ys : yc2-coordinate from C1/2 to shear center [m]. The shear center is the point where external forces only contributes to pure bending and no torsion.
    - E : modulus of elasticity [N/m2]
    - G : shear modulus of elasticity [N/m2]
    - Ix : area moment of inertia with respect to principal bending xe axis [m4]. This is the principal bending axis most parallel to the xc2 axis
    - Iy : area moment of inertia with respect to principal bending ye axis [m4]
    - K : torsional stiffness constant with respect to ze axis at the shear center [m4/rad]. For a circular section only this is identical to the polar moment of inertia.
    - k_x : shear factor for force in principal bending xe direction [-]
    - k_y : shear factor for force in principal bending ye direction [-]
    - A : cross sectional area [m2]
    - pitch : structural pitch about z_c2 axis. This is the angle between the xc2 -axis defined with the c2_def command and the main principal bending axis xe.
    - xe : xc2-coordinate from C1/2 to center of elasticity [m]. The elastic center is the point where radial force (in the z-direction) does not contribute to bending around the x or y directions.
    - ye : yc2-coordinate from C1/2 to center of elasticity [m]. The elastic center is the point where radial force (in the

    The autogenerated methods have the following structure

    def xxx(radius=None, mset=1, set=1):
        Parameters:
        -----------
        radius : int, float, array_like or None, optional
            Radius/radii of interest\n
            If int, float or array_like: values are interpolated to requested radius/radii
            If None (default): Values of all radii specified in st file returned
        mset : int, optional
            Main set number
        set : int, optional
            Sub set number


    Examples
    --------
    >>> stfile = StFile(r"tests/test_files/DTU_10MW_RWT_Blade_st.dat")
    >>> print (stfile.m()) # Mass at nodes
    [ 1189.51054664  1191.64291781  1202.76694262  ... 15.42438683]
    >>> print (st.E(radius=36, mset=1, set=1))  # Elasticity interpolated to radius 36m
    8722924514.652649
    >>> print (st.E(radius=36, mset=1, set=2))  # Same for stiff blade set
    8.722924514652648e+17
    """

    def __init__(self, filename=None):

        # in case the user wants to create a new non-existing st file
        if filename is None:
            self.main_data_sets = {}
            return

        with open(filename) as fid:
            txt = fid.read()
#         Some files starts with first set ("#1...") with out specifying number of sets
#         no_maindata_sets = int(txt.strip()[0])
#         assert no_maindata_sets == txt.count("#")
        self.main_data_sets = {}
        for mset in txt.split("#")[1:]:
            mset_nr = int(mset.strip().split()[0])
            set_data_dict = {}

            for set_txt in mset.split("$")[1:]:
                set_lines = set_txt.split("\n")
                set_nr, no_rows = map(int, set_lines[0].split()[:2])
                assert set_nr not in set_data_dict
                try:
                    # HAWC2 will ignore everything after the 19th element,
                    # some users have placed comments here after a ;
                    linelst = [set_lines[i].split(';')[0].split() for i in range(1, no_rows + 1)]
                except Exception as e:
                    print('it went wrong at (set/subset):', mset_nr, set_nr,
                          'with', no_rows, 'rows')
                    raise e
                set_data_dict[set_nr] = np.array(linelst, dtype=float)
            self.main_data_sets[mset_nr] = set_data_dict

        if len(linelst[0])==len(stc.split()):
            self.cols = stc.split()
        elif len(linelst[0])==len(fpm.split()):
            self.cols = fpm.split()
        else:
            raise TypeError('wrong number of columns in st file')

        for i, name in enumerate(self.cols):
            setattr(self, name, lambda radius=None, mset=1, set=1,
                    column=i: self._value(radius, column, mset, set))

    def _value(self, radius, column, mset_nr=1, set_nr=1):
        st_data = self.main_data_sets[mset_nr][set_nr]
        if radius is None:
            radius = self.radius_st(None, mset_nr, set_nr)
        return np.interp(radius, st_data[:, 0], st_data[:, column])

    def radius_st(self, radius=None, mset=1, set=1):
        r = self.main_data_sets[mset][set][:, 0]
        if radius is None:
            return r
        return r[np.argmin(np.abs(r - radius))]

    def to_str(self, mset=1, set=1, precision='%12.5e '):
        d = self.main_data_sets[mset][set]
        return '\n'.join([(precision * d.shape[1]) % tuple(row) for row in d])

    def set_value(self, mset_nr, set_nr, **kwargs):
        for k, v in kwargs.items():
            column = self.cols.index(k)
            self.main_data_sets[mset_nr][set_nr][:, column] = v

    def save(self, filename, precision='%15.07e', encoding='utf-8'):
        """Save all data defined in main_data_sets to st file.
        """

        # when creating empty st object, cols is not set yet
        if not hasattr(self, 'cols'):
            if self.main_data_sets[1][1].shape[1]==19:
                self.cols = stc.split()
            elif self.main_data_sets[1][1].shape[1]==30:
                self.cols = fpm.split()
            else:
                c = self.main_data_sets[1][1].shape[1]
                raise TypeError(f'St input needs 19 (iso) or 30 (aniso/fpm) cols, not {c}')

        colwidth = len(precision % 1)
        sep = '=' * colwidth * len(self.cols) + '\n'
        colhead = ''.join([k.center(colwidth) for k in self.cols]) + '\n'

        nsets = len(self.main_data_sets)
        # fails if dirname is empty string
        if len(os.path.dirname(filename)) > 0:
            os.makedirs(os.path.dirname(filename), exist_ok=True)
        with open(filename, 'w', encoding=encoding) as fid:
            fid.write('%i ; number of sets, Nset\n' % nsets)
            for mset, set_data_dict in self.main_data_sets.items():
                fid.write('#%i ; set number\n' % mset)
                for set, set_array in set_data_dict.items():
                    dstr = self.to_str(mset=mset, set=set, precision=precision)
                    npoints = self.main_data_sets[mset][set].shape[0]
                    fid.write(sep + colhead + sep)
                    fid.write('$%i %i\n' % (set, npoints))
                    fid.write(dstr + '\n')

    def element_stiffnessmatrix(self, radius, mset_nr, set_nr, length):
        """Compute the element stiffness matrix

        Parameters
        ----------
        radius : float
            radius of element (used of obtain element properties
        length : float
            eleement length
        """

        # not supported for FPM format
        if len(self.cols)==30:
            return

        K = np.zeros((13, 13))
        "r m x_cg y_cg ri_x ri_y x_sh y_sh E G I_x I_y I_p k_x k_y A pitch x_e y_e"
        ES1, ES2, EMOD, GMOD, IX, IY, IZ, KX, KY, A = [getattr(self, n)(radius, mset_nr, set_nr)
                                                       for n in "x_sh,y_sh,E,G,I_x,I_y,I_p,k_x,k_y,A".split(",")]
        ELLGTH = length

        ETAX = EMOD * IX / (KY * GMOD * A * ELLGTH**2)
        ETAY = EMOD * IY / (KX * GMOD * A * ELLGTH**2)
        ROX = 1 / (1 + 12 * ETAX)
        ROY = 1 / (1 + 12 * ETAY)
        ROY = 1 / (1 + 12 * ETAX)
        K[1, 1] = 12 * EMOD * IY * ROY / ELLGTH**3
        K[1, 5] = 6 * EMOD * IY * ROY / ELLGTH**2
        K[1, 6] = -K[1, 1] * ES2
        K[1, 7] = -K[1, 1]
        K[1, 11] = K[1, 5]
        K[1, 12] = -K[1, 6]
        K[2, 2] = 12 * EMOD * IX * ROX / ELLGTH**3
        K[2, 4] = -6 * EMOD * IX * ROX / ELLGTH**2
        K[2, 6] = K[2, 2] * ES1
        K[2, 8] = -K[2, 2]
        K[2, 10] = K[2, 4]
        K[2, 12] = -K[2, 6]
        K[3, 3] = A * EMOD / ELLGTH
        K[3, 9] = -K[3, 3]
        K[4, 4] = 4 * EMOD * IX * (1 + 3 * ETAX) * ROX / ELLGTH
        K[4, 6] = K[2, 4] * ES1
        K[4, 8] = -K[2, 4]
        K[4, 10] = 2 * EMOD * IX * (1 - 6 * ETAX) * ROX / ELLGTH
        K[4, 12] = -K[4, 6]
        K[5, 5] = 4 * EMOD * IY * (1 + 3 * ETAY) * ROY / ELLGTH
        K[5, 6] = -K[1, 5] * ES2
        K[5, 7] = -K[1, 5]
        K[5, 11] = 2 * EMOD * IY * (1 - 6 * ETAY) * ROY / ELLGTH
        K[5, 12] = -K[5, 6]
        K[6, 6] = GMOD * IZ / ELLGTH + 12 * EMOD * (IX * ES1**2 * ROX + IY * ES2**2 * ROY) / ELLGTH**3
        K[6, 7] = K[1, 12]
        K[6, 8] = K[2, 12]
        K[6, 10] = -K[4, 12]
        K[6, 11] = -K[5, 12]
        K[6, 12] = -K[6, 6]
        K[7, 7] = K[1, 1]
        K[7, 11] = -K[1, 5]
        K[7, 12] = K[1, 6]
        K[8, 8] = K[2, 2]
        K[8, 10] = -K[2, 4]
        K[8, 12] = K[2, 6]
        K[9, 9] = K[3, 3]
        K[10, 10] = K[4, 4]
        K[10, 12] = -K[4, 6]
        K[11, 11] = K[5, 5]
        K[11, 12] = -K[5, 6]
        K[12, 12] = K[6, 6]
        K = K[1:, 1:]
        K = K + K.T - np.eye(12) * K
        return K

    def shape_function_ori(self, radius, mset_nr, set_nr, length, z):

        # not supported for FPM format
        if len(self.cols)==30:
            return

        XSC, YSC, EMOD, GMOD, IX, IY, IZ, KX, KY, AREA = [getattr(self, n)(radius, mset_nr, set_nr)
                                                          for n in "x_sh,y_sh,E,G,I_x,I_y,I_p,k_x,k_y,A".split(",")]

        etax = EMOD * IX / KY / GMOD / AREA / (length**2)
        etay = EMOD * IY / KX / GMOD / AREA / (length**2)
        rhox = 1 / (1 + 12 * etax)
        rhoy = 1 / (1 + 12 * etay)

        f1 = z / length
        f2 = 1 - f1
        f3x = f1 * (3 * f1 - 2 * f1**2 + 12 * etax) * rhox
        f4x = f2 * (3 * f2 - 2 * f2**2 + 12 * etax) * rhox
        f5x = length * (f1**2 - (1 - 6 * etax) * f1 - 6 * etax) * f1 * rhox
        f6x = length * (f2**2 - (1 - 6 * etax) * f2 - 6 * etax) * f2 * rhox
        f7x = 6 / length * f1 * f2 * rhox
        f8x = f1 * (3 * f1 - 2 * (1 - 6 * etax)) * rhox
        f9x = f2 * (3 * f2 - 2 * (1 - 6 * etax)) * rhox

        f3y = f1 * (3 * f1 - 2 * f1**2 + 12 * etay) * rhoy
        f4y = f2 * (3 * f2 - 2 * f2**2 + 12 * etay) * rhoy
        f5y = length * (f1**2 - (1 - 6 * etay) * f1 - 6 * etay) * f1 * rhoy
        f6y = length * (f2**2 - (1 - 6 * etay) * f2 - 6 * etay) * f2 * rhoy
        f7y = 6 / length * f1 * f2 * rhoy
        f8y = f1 * (3 * f1 - 2 * (1 - 6 * etay)) * rhoy
        f9y = f2 * (3 * f2 - 2 * (1 - 6 * etay)) * rhoy

        return np.array([[f4y, 0, 0, 0, -f7y, 0],
                         [0,  f4x, 0,  f7x, 0, 0],
                         [0, 0,   f2, 0, 0, 0],
                         [0, f6x, 0, f9x, 0, 0],
                         [-f6y, 0, 0, 0, f9y, 0],
                         [(f2 - f4y) * YSC, -(f2 - f4x) * XSC, 0, f7x * XSC, f7y * YSC, f2],
                         [f3y, 0, 0, 0, f7y, 0],
                         [0, f3x, 0, -f7x, 0, 0],
                         [0, 0,    f1, 0, 0, 0],
                         [0, -f5x, 0, f8x, 0, 0],
                         [f5y, 0, 0, 0, f8y, 0],
                         [(f1 - f3y) * YSC, -(f1 - f3x) * XSC, 0, -f7x * XSC, -f7y * YSC, f1]]).T


if __name__ == "__main__":
    import os
    cwd = os.path.dirname(__file__)
    st = StFile(os.path.join(cwd, r'tests/test_files/DTU_10MW_RWT_Blade_st.dat'))
    print(st.m())
    print(st.E(radius=36, mset=1, set=1))  # Elastic blade
    print(st.E(radius=36, mset=1, set=2))  # stiff blade
    #print (st.radius())
    xyz = np.array([st.x_e(), st.y_e(), st.r()]).T[:40]
    n = 2
    xyz = np.array([st.x_e(None, 1, n), st.y_e(None, 1, n), st.r(None, 1, n)]).T[:40]
    #print (xyz)
    print(np.sqrt(np.sum((xyz[1:] - xyz[:-1]) ** 2, 1)).sum())
    print(xyz[-1, 2])
    print(np.sqrt(np.sum((xyz[1:] - xyz[:-1]) ** 2, 1)).sum() - xyz[-1, 2])
    print(st.x_e(67.8883), st.y_e(67.8883))
    #print (np.sqrt(np.sum(np.diff(xyz, 0) ** 2, 1)))
    print(st.pitch(67.8883 - 0.01687))
    print(st.pitch(23.2446))

    # print (st.)
    # print (st.)



import pandas as pd
import numpy as np
import re, os
import click
from scipy import signal
from functools import wraps 
from wetb.hawc2.Hawc2io import ReadHawc2
from wetb.fatigue_tools.fatigue import eq_load

    
def isFloat(string):
    try:
        float(string)
        return True
    except ValueError:
        return False


class HAWC2DataFrame(pd.DataFrame):
    """
    A subclass of a pandas.DataFrame. The HAWC2DataFrame is able to link to a
    directory of HAWC2 result files. The HAWC2DataFrame contains an Accessor,
    named 'wetb', which contains functions to  populate the HAWC2DataFrame with
    post-processed statistics, and to aggregate the statistics.
    """

    _metadata = ['_fields', 'channels', '_directory', '_filenames']
    
    @property
    def _constructor(self):
        return HAWC2DataFrame
        
        
    def __init__(self, *args, dir=None, pattern=None, channels=None, **kwargs):
        
        if all(x is not None for x in [dir, channels]):            
            if pattern is None:
                res = self.link_directory(dir, channels)
            else:
                res = self.link_results(dir, pattern, channels)
            super(HAWC2DataFrame, self).__init__(res)

        else:
            super(HAWC2DataFrame, self).__init__(*args, **kwargs)


    
    def __finalize__(self, other, method=None, **kwargs):
        """propagate metadata from other to self """
        # merge operation: using metadata of the left object
        if method == 'merge':
            for name in self._metadata:
                object.__setattr__(self, name, getattr(other.left, name, None))
        # concat operation: using metadata of the first object
        elif method == 'concat':
            for name in self._metadata:
                object.__setattr__(self, name, getattr(other.objs[0], name, None))
        else:
            for name in self._metadata:
                object.__setattr__(self, name, getattr(other, name, None))
        return self

    
    def link_results(self, directory, pattern_string, channels):
        self._directory = directory
        pattern, self._fields = self._compile_pattern(pattern_string)

        # get all filenames that fit the pattern
        self._filenames = [x[:-4] for x in os.listdir(directory) if x.endswith('.sel')]
        self._filenames = [x for x in self._filenames if pattern.match(x)]
        
        self.channels = channels
        # Extract input attributes and put in dataframe
        dat = []
        for fn in self._filenames:
            dat.append([float(x) if isFloat(x) else x for x in pattern.findall(fn)[0]])
        dat = pd.DataFrame(dat)
        
        # set column multi index
        if not dat.empty:
            column_tuples = list(zip(*[self._fields, ['']*len(self._fields)]))
            dat.columns = pd.MultiIndex.from_tuples(column_tuples, names=['channel', 'stat'])
        return dat
    
    
    def link_directory(self, directory, channels):
        '''
        Links all result files in a directory. Used if no pattern is given.
        '''
        self._directory = directory
        self.channels = channels

        self._fields = ['filename']
        # get all filenames of result files
        self._filenames = [x[:-4] for x in os.listdir(directory) if x.endswith('.sel')]
        dat = pd.DataFrame(self._filenames)

        # set column multi index
        if not dat.empty:
            column_tuples = list(zip(*[self._fields, ['']*len(self._fields)]))
            dat.columns = pd.MultiIndex.from_tuples(column_tuples, names=['channel', 'stat'])
        return dat
        
            
    @staticmethod
    def _compile_pattern(pattern_string):
        brackets = re.compile('{(.*?)}')
        fields = brackets.findall(pattern_string)
        for field in fields:
            pattern_string = pattern_string.replace('{'+ field +'}', '(.*)')
        return re.compile(pattern_string), fields   
        
                 
    def __call__(self, **kwargs):
        return self[self._mask( **kwargs)]


    def _mask(self, **kwargs):
        '''
        Returns a mask for refering to a dataframe, or self.Data, or self.Data_f, etc.
        example. dlc.mask(wsp=[12, 14], controller='noIPC')
        '''
        N = len(self)
        mask = [True] * N
        for key, value in kwargs.items():
            if isinstance(value, (list, tuple, np.ndarray)):
                mask_temp = [False] * N
                for v in value:
                    mask_temp = mask_temp | (self[key] == v)
                mask = mask & mask_temp
            else: #scalar, or single value
                mask = mask & (self[key] == value)
        return mask



@pd.api.extensions.register_dataframe_accessor("wetb")
class wetbAccessor(object):
    def __init__(self, pandas_obj):
        self._validate(pandas_obj)
        self._obj = pandas_obj
    
        
    @staticmethod
    def _validate(obj):
        # verify that a HAWC2DataFrame is using the wetb accessor.
        if not isinstance(obj, HAWC2DataFrame):
            raise TypeError(f"wetb namespace is not accessible by type {type(obj)}. Expected: { type(HAWC2DataFrame())}")
            
            
    @classmethod    
    def populate_method(cls, method_name, label=None):
        label = label or method_name
        def decorator(func):
            @wraps(func) 
            def wrapper(self, channels, *args, **kwargs): 
                return self._add_stat(func, label, channels, *args, **kwargs)
            setattr(cls, method_name, wrapper)
            # Note we are not binding func, but wrapper which accepts self but does exactly the same as func
            return func # returning func means func can still be used normally
        return decorator   
        
                 
    def fetch_result(self, idx, channels=None):
        '''
        Returns a time series dataframe of a single result file given an index number.
        '''
        fn = self._obj._filenames[idx]
        channels = channels or self.channels

        raw = ReadHawc2(os.path.join(self._obj._directory, fn)).ReadAll(ChVec=[i-1 for i in channels.values()])
        #raw = readHawc2Res(os.path.join(self._directory, fn), channels)
        return pd.DataFrame(raw, columns=channels.keys())
        # except:
        #     print(f'File {fn} could not be loaded.')
        
            
    def iter_sim(self, **kwargs):
        '''
        Iterates over simulation result files given the input filter defined in kwargs.
        '''
        sims_to_iterate = self._obj(**kwargs)
        for idx, row in sims_to_iterate.iterrows():
            raw = self.fetch_result(idx)
            yield row[self._obj._fields], raw

    
    def _add_stat(self, func, stat_name, channels=None, *args, **kwargs):
        '''
        Adds a column of statistics for the given channels using the given function.
        The function should take a pandas series (1d array) and return a float.
        '''
        values = []
        if channels is None:
            channels = self._obj.channels
        else:
            channels = {k:v for k,v in self._obj.channels.items() if k in channels}
        
        channel_string = ', '.join(channels)
        print(f'Calculating {stat_name} for {channel_string}...')
        
        with click.progressbar(self._obj.index) as bar:
            for idx in bar:
                raw = self.fetch_result(idx, channels)
                values.append(func(raw, *args, **kwargs))

        df = pd.DataFrame(values)
        # add multi index columns
        col_ch     = list(channels.keys())
        col_stat   = [stat_name]*len(col_ch)
        col_tuples = list(zip(*[col_ch, col_stat]))
        df.columns = pd.MultiIndex.from_tuples(col_tuples, names=['channel', 'stat'])

        return self._obj.join(df)


    def aggregate_rows(self, key):
        # used for taking the mean over all seeds
        print(f'Calculating mean over key={key}...')
        in_fields = [x for x in self._obj._fields if x != key]
        out_fields = [x for x in list(self._obj) if x[0] not in self._obj._fields]
        in_atts = self._obj[in_fields].drop_duplicates()
        new_dat = []
        with click.progressbar(in_atts.iterrows()) as bar:
            for _, x in bar:
                filt = {k[0]: v for k, v in dict(x).items()}
                new_dat.append(list(x) + list(self._obj(**filt)[out_fields].mean().values))
            
        # Make new HAWC2DataFrame and copy metadata (is there a better way?)
        new_df = HAWC2DataFrame(new_dat)
        for attr in self._obj._metadata:
            setattr(new_df, attr, getattr(self._obj, attr))
        new_df._fields = in_fields

        # add multi index columns
        col_ch     = in_fields + [x[0] for x in out_fields]
        col_stat   = ['']*len(in_fields) + [x[1] for x in out_fields]
        col_tuples = list(zip(*[col_ch, col_stat]))
        new_df.columns = pd.MultiIndex.from_tuples(col_tuples,
                    names=['channel', 'stat'])
        return new_df


    def aggregate_columns(self, source, dest, drop=False):
        # used for taking the mean over all blades.
        channel_string = ', '.join(source)
        print(f'Calculating mean over {channel_string}...')
        new_df = self._obj.copy()
        # get all unique stat indices
        col_stat = list(set(x[1] for x in list(self._obj)))
        for stat in col_stat:
            keys = [x for x in list(self._obj) if x[0] in source and x[1] == stat]
            if not keys:
                continue
            new_df[(dest, stat)] = self._obj[keys].mean(axis=1)
            
        if drop:
            # delete source columns
            new_df = new_df.drop(columns=source)
        return new_df

    
    @staticmethod
    def read_csv(filename):
        '''
        Reads a postproc csv file which was generated by the HAWC2Res class. Returns
        a DataFrame
        '''
        df           = HAWC2DataFrame(pd.read_csv(filename, header =[0, 1]))
        multicol     = list(df)
        multicol_new = []
        for col in multicol:
            if 'Unnamed' in col[1]:
                multicol_new.append((col[0], ''))
            else:
                multicol_new.append(col)
        df.columns = pd.MultiIndex.from_tuples(multicol_new)
        return df


@wetbAccessor.populate_method('DEL')
def DEL(x, m=4):
    '''
    Calculates the damage equivalent load of a set of time series loads.
    args:
        x (pd.DataFrame): load time series, where each column is a different load channel
        m (float): wohler constant (default: 4)
    returns:
        DEL (list of floats): Damage equivalent load for each load channel.
    '''
    DEL = []
    for k in list(x):
        DEL.append(eq_load(x[k].values, m=m)[0][0])
    return DEL
    
    
@wetbAccessor.populate_method('mean', label='Mean')
def _mean(x):
    '''
    Calculates the mean of a set of time series.
    args:
        x (pd.DataFrame): time series, where each column contains a different time series.
    returns:
        (list of floats): Mean of each time series.
    '''
    return x.mean().values
    
    
@wetbAccessor.populate_method('var', label='Var')
def _var(x):
    '''
    Calculates the variance of a set of time series.
    args:
        x (pd.DataFrame): time series, where each column contains a different time series.
    returns:
        (list of floats): variance of each time series.
    '''
    return x.var().values


@wetbAccessor.populate_method('std', label='Std')
def _std(x):
    '''
    Calculates the standard deviation of a set of time series.
    args:
        x (pd.DataFrame): time series, where each column contains a different time series.
    returns:
        (list of floats): standard deviation of each time series.
    '''
    return x.std().values


@wetbAccessor.populate_method('final')
def _final(x):
    '''
    Returns the final value of a set of time series.
    args:
        x (pd.DataFrame): time series, where each column contains a different time series.
    returns:
        (list of floats): final value of each time series.
    '''
    return x.iloc[-1].values


@wetbAccessor.populate_method('PSD')
def _psd(x, fs=100):
    '''
    Calculates power spectral density (PSD) of a set of time series.
    args:
        x (pd.DataFrame): time series, where each column contains a different time series.
        fs (float): sampling frequency in Hz (default: 100Hz).
    returns:
        (list of arrays): PSD of each time series.
    '''
    PSD = []
    for k in list(x):
        f, Py = signal.welch(x[k].values, fs=fs, nperseg=1024*8)
        PSD.append(Py)
    return PSD
            



# -*- coding: utf-8 -*-
"""
Created on Mon Jun 18 13:34:12 2018

@author: shfe

Note:
This script is used for Klinkvort model used for cyclic accumulation calculation
"""

import numpy as np
import matplotlib.pyplot as plt
plt.close('all')

class KlinkvortModel(object):
    
    
    def __init__(self, xi_b, xi_c, N):
        """
        Initialize the input parameters
        """
        
        self.xi_b = xi_b
        self.xi_c = xi_c
        self.N = N
        
    def alpha(self):
        """
        Calculate the parameter alpha
        """
        # Calculate the Tb, Tc
        if not 0 <= self.xi_b <= 1:
            raise ValueError('The maximum load ratio should between 0 to 1')
        
        if not -1 <= self.xi_c <= 1:
            raise ValueError('The cyclic load ratio should between -1 to 1')
            
        if self.xi_b <= 0.022:
            Tb = 0
        else:
            Tb = 0.61*self.xi_b - 0.013
        
        Tc = (self.xi_c + 0.63)*(self.xi_c - 1)*(self.xi_c - 1.64)
        alpha = Tb * Tc
        
        return Tb, Tc, alpha 
    
    def evoluation(self, y1):
        """
        Evoluation model for cyclic accumulated rotation
        """
        
        _, _, alpha = self.alpha()
        yn = y1 * self.N**alpha
        
        return yn
        
    def equivalentnum(self, y1, yn, alpha):
        """
        Equivalent number for cylces
        """
        num = (yn/y1)**(1/alpha)
        
        return num
        
def superposition(xi_b, xi_c, N_cases, y1_cases, s0 = 0):

    
    # Check the array equal number of items
    if not (np.size(xi_b) == np.size(xi_c) and np.size(xi_b) == np.size(N_cases)):
        raise ValueError('Size for input should be identical')
    print(s0)
    yn = np.zeros(np.size(xi_b))
    for i, (b, c, N, y1) in enumerate(zip(xi_b, xi_c, N_cases, y1_cases)):
        
        if i==0:
            
            em = KlinkvortModel(b, c, N)
            _, _, alpha = em.alpha()
            if s0 == 0:
                yn[i] = em.evoluation(y1)
            else:
                N_eq = (s0/y1)**(1/alpha)
                if N_eq < 1e8:
                    
#                    N_eq = (s0/y1)**(1/alpha)
                    yn[i] = y1*(N+N_eq)**alpha
                else:
                    yn[i] = s0
                        
                
        
        else:
            em = KlinkvortModel(b, c, N)
            _, _, alpha = em.alpha()
            N_eq = (yn[i-1]/y1)**(1/alpha)
            if N_eq < 1e8:
#                N_eq = (yn[i-1]/y1)**(1/alpha)
                yn[i] = y1*(N_eq + N)**alpha
            else:
                yn[i] = yn[i-1]

    return yn
    
def alpha_super(xi_b, xi_c, N_cases):
    # Check the alpha value
    if not (np.size(xi_b) == np.size(xi_c) and np.size(xi_b) == np.size(N_cases)):
        raise ValueError('Size for input should be identical')
    
    alpha_list = np.zeros(np.size(xi_b))
    for i, (b, c, N) in enumerate(zip(xi_b, xi_c, N_cases)):
        
        em = KlinkvortModel(b, c, N)
        _, _, alpha = em.alpha()
        alpha_list[i] = alpha
        
    return alpha_list
    
    
    
    


d = None
d = dir()

from .htc_file import HTCFile
from .log_file import LogFile
from .ae_file import AEFile
from .at_time_file import AtTimeFile
from .pc_file import PCFile
from . import shear_file
from .st_file import StFile
from .hawc2_input_writer import HAWC2InputWriter

__all__ = sorted([m for m in set(dir()) - set(d)])



# -*- coding: utf-8 -*-
"""
Created on Wed Nov 23 11:22:50 2011

@author: dave
"""
# external libraries
import numpy as np

import matplotlib as mpl
from matplotlib.figure import Figure
# use a headless backend
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigCanvas
# wafo is an optional dependency only required for non default PSD peak marking
try:
    import wafo
except ImportError:
    pass


def make_fig(nrows=1, ncols=1, figsize=(12,8), dpi=120):
    """

    Equivalent function of pyplot.subplots(). The difference is that this one
    is not interactive and is used with backend plotting only.

    Parameters
    ----------
    nrows=1, ncols=1, figsize=(12,8), dpi=120

    Returns
    -------
    fig, canvas, axes


    """
    return subplots(nrows=nrows, ncols=ncols, figsize=figsize, dpi=dpi)


def subplots(nrows=1, ncols=1, figsize=(12,8), dpi=120, num=0, subplot_kw={},
             gridspec_kw={}):
    """

    Equivalent function of pyplot.subplots(). The difference is that this one
    is not interactive and is used with backend plotting only.

    Parameters
    ----------
    nrows=1, ncols=1, figsize=(12,8), dpi=120

    num : dummy variable for compatibility

    Returns
    -------
    fig, axes


    """

    fig = Figure(figsize=figsize, dpi=dpi)
    axes = fig.subplots(nrows=nrows, ncols=ncols,
                        subplot_kw=subplot_kw, gridspec_kw=gridspec_kw)
    try:
        axes = axes.reshape((nrows,ncols))
    except AttributeError:
        pass
    # canvas = FigCanvas(fig)
    # fig.set_canvas(canvas)
    # axes = np.ndarray((nrows, ncols), dtype=object)
    # plt_nr = 1
    # for row in range(nrows):
    #     for col in range(ncols):
    #         axes[row,col] = fig.add_subplot(nrows, ncols, plt_nr, **subplot_kw)
    #         plt_nr += 1
    return fig, axes


def match_axis_ticks(ax1, ax2, ax1_format=None, ax2_format=None):
    """
    Match ticks of ax2 to ax1
    =========================

    ax1_format: '%1.1f'

    Parameters
    ----------

    ax1, ax2

    Returns
    -------

    ax1, ax2

    """
    # match the ticks of ax2 to ax1
    yticks1 = len(ax1.get_yticks())
    ylim2 = ax2.get_ylim()
    yticks2 = np.linspace(ylim2[0], ylim2[1], num=yticks1).tolist()
    ax2.yaxis.set_ticks(yticks2)

    # give the tick labels a given precision
    if ax1_format:
        majorFormatter = mpl.ticker.FormatStrFormatter(ax1_format)
        ax1.yaxis.set_major_formatter(majorFormatter)

    if ax2_format:
        majorFormatter = mpl.ticker.FormatStrFormatter(ax2_format)
        ax2.yaxis.set_major_formatter(majorFormatter)

    return ax1, ax2


def one_legend(*args, **kwargs):
    """First list all the axes as arguments. Any keyword arguments will be
    passed on to ax.legend(). Legend will be placed on the last axes that was
    passed as an argument.

    Parameters
    ----------

    Returns
    -------

    legend

    """
    # or more general: not only simple line plots (bars, hist, ...)
    objs = []
    for ax in args:
        objs += ax.get_legend_handles_labels()[0]
#    objs = [ax.get_legend_handles_labels()[0] for ax in args]
    labels = [obj.get_label() for obj in objs]
    # place the legend on the last axes
    leg = ax.legend(objs, labels, **kwargs)
    return leg


def p4psd(ax, rpm_mean, fmax=10, y_pos_rel=0.25, color='g', ls='--', ps=None,
          col_text='w'):
    """
    Add the P's on a PSD

    fmax is the maximum value on the plot (ax.xlim). This only works when
    setting the xlim of the plot before calling p4psd.

    Parameters
    ----------

    ax

    rpm_mean

    fmax : int, default=17
        stop plotting p's after fmax (when ps is None)

    ps : iterable of ints, default=None
        specify which p's to plot (ignores fmax)

    y_pos_rel : int or list, default=0.25
    """

    if ps is None:
        pmax = int(60*fmax/rpm_mean)
        ps = list(range(1, pmax))
    else:
        pmax = len(ps)

    if isinstance(y_pos_rel, float) or isinstance(y_pos_rel, int):
        y_pos_rel = [y_pos_rel]*pmax

    f_min = ax.get_xlim()[0]
    f_max = ax.get_xlim()[1]

    # add the P's
    bbox = dict(boxstyle="round", edgecolor=color, facecolor=color)
    for i, p in enumerate(ps):
        p_freq = p * rpm_mean / 60.0
        if p%3 == 0:
            alpha=0.5
            ax.axvline(x=p_freq, linewidth=1, color=color, alpha=0.7, ls=ls)
        else:
            alpha = 0.2
            ax.axvline(x=p_freq, linewidth=1, color=color, alpha=0.7, ls=ls)

        x = (p_freq - f_min) / (f_max - f_min)
        y = y_pos_rel[i]

        p_str = '%iP' % p
        bbox['alpha'] = alpha
        ax.text(x, y, p_str, fontsize=8, verticalalignment='bottom',
                horizontalalignment='center', bbox=bbox, color=col_text,
                transform=ax.transAxes)

    return ax


def peaks(ax, freqs, Pxx, fn_max, min_h, nr_peaks=15, min_p=0, col_line='k',
          ypos_mean=0.14, col_text='w', ypos_delta=0.06, bbox_alpha=0.5,
          verbose=False, format_text='%2.2f', period=False):
    """
    indicate the peaks

    Parameters
    ----------

    period : boolean, default=False
        If True, the period instead of the frequency is given in the text

    min_h : float
        The threshold in the rainflowfilter (default 0.05*range(S(:))).
        A zero value will return all the peaks of S.

    min_p : float, 0..1
        Only the peaks that are higher than min_p*max(max(S))
        min_p*(the largest peak in S) are returned (default  0).
    """
    i_fn_max = np.abs(freqs - fn_max).argmin()
    # ignore everything above fn_max
    freqs = freqs[:i_fn_max]
    Pxx = Pxx[:i_fn_max]
    Pxx_log = 10.*np.log10(Pxx)
    try:
        pi = wafo.misc.findpeaks(Pxx_log, n=len(Pxx), min_h=min_h, min_p=min_p)
        if verbose:
            print('len Pxx', len(Pxx_log), 'nr of peaks:', len(pi))
    except Exception as e:
        if verbose:
            print('len Pxx', len(Pxx_log))
            print('*** wafo.misc.findpeaks FAILED ***')
            print(e)
        return ax

    # only take the nr_peaks most significant heights
    pi = pi[:nr_peaks]
    # and sort them accoriding to frequency (or index)
    pi.sort()

    # mark the peaks with a circle
#    ax.plot(freqs[pi], Pxx[:xlim][pi], 'o')
    # and mark all peaks
    switch = True
    # yrange_plot = Pxx_log.max() - Pxx_log.min()
    for peak_nr, ii in enumerate(pi):
        freq_peak = freqs[ii]
#        Pxx_peak = Pxx_log[ii]
        # take the average frequency and plot vertical line
        ax.axvline(x=freq_peak, linewidth=1, color=col_line, alpha=0.6)
        if switch:
            # locate at the min value (down the plot), but a little
            # lower so it does not interfere with the plot itself
            # if ax.set_yscale('log') True, set log values as coordinates!
            # text_ypos = Pxx_log.min() + yrange_plot*0.1
            text_ypos = ypos_mean + ypos_delta
            switch = False
        else:
            # put it a little lower than the max value so it does
            # not mess with the title (up the plot)
            # text_ypos = Pxx_log.min() - yrange_plot*0.4
            text_ypos = ypos_mean - ypos_delta
            switch = True
#        print('%2.2e %2.2e %2.2e' % (yrange_plot, Pxx[:xlim].max(), Pxx[:xlim].min())
#        print peak, text_ypos
#        print textbox
#        print yrange_plot
        xrel = freq_peak/fn_max
        # and the value in a text box
        if period:
            freq_peak = 1/freq_peak
        textbox = format_text % freq_peak
        ax.text(xrel, text_ypos, textbox, fontsize=10, transform=ax.transAxes,
                 va='bottom', color=col_text, bbox=dict(boxstyle="round",
                 ec=col_line, fc=col_line, alpha=bbox_alpha,))

    return ax


def match_yticks(ax1, ax2, nr_ticks_forced=None, extend=False):
    """
    """

    if nr_ticks_forced is None:
        nr_yticks1 = len(ax1.get_yticks())
    else:
        nr_yticks1 = nr_ticks_forced
        ylim1 = ax1.get_ylim()
        yticks1 = np.linspace(ylim1[0], ylim1[1], num=nr_yticks1).tolist()
        ax1.yaxis.set_ticks(yticks1)

    ylim2 = ax2.get_ylim()
    yticks2 = np.linspace(ylim2[0], ylim2[1], num=nr_yticks1).tolist()
    ax2.yaxis.set_ticks(yticks2)

    if extend:
        offset1 = (ylim1[1] - ylim1[0])*0.1
        ax1.set_ylim(ylim1[0]-offset1, ylim1[1]+offset1)
        offset2 = (ylim2[1] - ylim2[0])*0.1
        ax2.set_ylim(ylim2[0]-offset2, ylim2[1]+offset2)

    return ax1, ax2


def psd(ax, time, sig, nfft=None, res_param=250, f0=0, f1=None, nr_peaks=10,
        min_h=15, min_p=0, mark_peaks=False, col='r-', label=None, alpha=1.0,
        ypos_peaks=0.9, ypos_peaks_delta=0.12, format_text='%2.2f', period=False):
    """Only plot the psd on a given axis and optionally mark the peaks.
    """

    sps = int(round(1.0/np.diff(time).mean(), 0))
    if f1 is None:
        f1 = sps/2.0

    if nfft is None:
        nfft = int(round(res_param * sps / (f1-f0), 0))
    if nfft > len(sig):
        nfft = len(sig)

    # calculate the PSD

    Pxx, freqs = mpl.mlab.psd(sig, NFFT=nfft, Fs=sps) # window=None
    # for clean signals (steps, sinus) use another window
    # for stochastic signal, the default window Hanning is good
    # in Scipy you can provide strings for the different windows, not so in mlab

    i0 = np.abs(freqs - f0).argmin()
    i1 = np.abs(freqs - f1).argmin()

    # plotting psd, marking peaks
    ax.plot(freqs[i0:i1], Pxx[i0:i1], col, label=label, alpha=alpha)
    if mark_peaks:
        ax = peaks(ax, freqs[i0:i1], Pxx[i0:i1], fn_max=f1, min_p=min_p,
                   nr_peaks=nr_peaks, col_line=col[:1], format_text=format_text,
                   ypos_delta=ypos_peaks_delta, bbox_alpha=0.5, period=period,
                   ypos_mean=ypos_peaks, min_h=min_h, col_text='w')

    return ax


def time_psd(results, labels, axes, alphas=[1.0, 0.7], colors=['k-', 'r-'],
             NFFT=None, res_param=250, f0=0, f1=None, nr_peaks=10, min_h=15,
             mark_peaks=False, xlabels=['frequency [Hz]', 'time [s]'],
             ypos_peaks=[0.3, 0.9], ypos_peaks_delta=[0.12, 0.12]):
    """
    Plot time series and the corresponding PSD of the channel.

    The frequency range depends on the sample rate: fn_max = sps/2.
    The number of points is: NFFT/2
    Consequently, the frequency resolution is: NFFT/SPS (points/Hz)

    len(data) > NFFT otherwise the results do not make any sense

    res_param = NFFT*frequency_range*sps
    good range for nondim_resolution is: 200-300

    Frequency range is based on f0 and f1

    for the PSD powers of 2 are faster, but the difference in performance with
    any other random number fluctuates on average and is not that relevant.

    Requires two axes!

    With the PSD peaks this only works nicely for 2 results

    Parameters
    ----------

    results : list
        list of time,data pairs (ndarrays)
    """

    axes = axes.ravel()

    for i, res in enumerate(results):
        time, data = res
        label = labels[i]
        col = colors[i]
        alpha = alphas[i]
        if isinstance(NFFT, list):
            nfft = NFFT[i]
        else:
            nfft = NFFT
        axes[0] = psd(axes[0], time, data, nfft=nfft, res_param=res_param,
                      f0=f0, f1=f1, nr_peaks=nr_peaks, min_h=min_h,
                      mark_peaks=mark_peaks, col=col, label=label, alpha=alpha,
                      ypos_peaks_delta=ypos_peaks_delta[i],
                      ypos_peaks=ypos_peaks[i])

        # plotting time series
        axes[1].plot(time, data, col, label=label, alpha=alpha)

    axes[0].set_yscale('log')
    if isinstance(xlabels, list):
        axes[0].set_xlabel(xlabels[0])
        axes[1].set_xlabel(xlabels[1])
    for ax in axes:
        leg = ax.legend(loc='best', borderaxespad=0)
        # leg is None when no labels have been defined
        if leg is not None:
            leg.get_frame().set_alpha(0.7)
        ax.grid(True)

    return axes


def get_list_colors(nr, cmap='magma'):
    """Returns a list of color tuples

    Paramters
    ---------

    nr : int
        number of colors

    cmap : str
        a matplotlib color map name, for example: viridis, plasma, magma, hot,
        cool, binary, see also https://matplotlib.org/users/colormaps.html

    Returns
    -------

    clist : list
        List continaing 'nr' color tuples (with 3 elements).

    """
    # define the number of positions you want to have the color for
    # select a color map
    cmap = mpl.cm.get_cmap(cmap, nr)
    # other color maps: 	cubehelix, afmhot, hot
    # convert to array
    cmap_arr = cmap(np.arange(nr))
    # and now you have each color as an RGB tuple as
    clist = []
    for i in cmap_arr:
        clist.append(tuple(i[0:3]))

    return clist


if __name__ == '__main__':

    pass



# -*- coding: utf-8 -*-
"""
Created on Sun Apr 22 17:06:06 2018

@author: shfe

Hyperbolic model used for soil load displacement behaviour
"""

import numpy as np

def hyperbolic_load(y, pu, a, b):
    """
    Static hyperbolic model developed by Lee
    
    Parameters
    ----------------
    y -- displacement
    pu -- ultimate load
    a, b -- coefficients
    
    Output
    ----------------
    p -- soil resistence pressure
    """
    
    p = y/(a + b*y) * pu
        
    return p
    
def hyperbolic_disp(p, pu, a, b):
    """
    Static hyperbolic model developed by Lee
    
    Parameters
    ----------------
    p -- load
    pu -- ultimate load
    a, b -- coefficients
    
    Output
    ----------------
    y -- displacement
    """
    
    y = p * a/(pu - p*b)
        
    return y


# -*- coding: utf-8 -*-
"""
Created on Thu May 24 13:41:56 2018

@author: dave
"""

import time
import os
from multiprocessing import Pool
from glob import glob
from argparse import ArgumentParser
import json

import numpy as np
import pandas as pd
# import scipy.io as sio
import tables as tbl

from tqdm import tqdm

from wetb.utils.envelope import compute_envelope
from wetb.prepost import windIO
from wetb.prepost.Simulations import EnvelopeClass
from wetb.prepost.simchunks import AppendDataFrames


def logcheck(fname):
    """Check the log file of a single HAWC2 simulation and save results to
    textfile.
    """
    fsave = None
    mode = 'w'

    logf = windIO.LogFile()
    logf.readlog(fname)
    contents = logf._msglistlog2csv('')
    if fsave is None:
        fsave = fname.replace('.log', '.csv')
    with open(fsave, mode) as f:
        f.write(contents)


def add_channels(res):
    """Add channels in memory so they can be included in the statistics calculations
    """

    # EXAMPLE: tower bottom resultant moment
    # chitx = res.ch_dict['tower-tower-node-001-momentvec-x']['chi']
    # chity = res.ch_dict['tower-tower-node-001-momentvec-y']['chi']
    # mx = res.sig[:,chitx]
    # my = res.sig[:,chity]
    # data = np.sqrt(mx*mx + my*my).reshape(len(res.sig),1)
    # # add the channel
    # res.add_channel(data, 'TB_res', 'kNm', 'Tower bottom resultant bending moment')

    # blade resultant loads
    chitx = res.ch_dict['blade1-blade1-node-004-forcevec-x']['chi']
    chity = res.ch_dict['blade1-blade1-node-004-forcevec-y']['chi']
    x = res.sig[:,chitx]
    y = res.sig[:,chity]
    data = np.sqrt(x*x + y*y).reshape(len(res.sig),1)
    res.add_channel(data, 'BR_F_res', 'kN', 'Blade root resultant shear force')

    chitx = res.ch_dict['blade1-blade1-node-004-momentvec-x']['chi']
    chity = res.ch_dict['blade1-blade1-node-004-momentvec-y']['chi']
    x = res.sig[:,chitx]
    y = res.sig[:,chity]
    data = np.sqrt(x*x + y*y).reshape(len(res.sig),1)
    res.add_channel(data, 'BR_M_res', 'kNm', 'Blade root resultant bending moment')

    return res


def loads_at_extreme(dfres, chans_selmax, chans_atmax):
    """For a range of channels defined in chans_list, list their value respective
    values at the time when a max/min occurs.

    It makes most sense to use chans_selmax == chans_atmax (BLADED style reports)

    Other use is to track loads at for example the maximum yaw angle.
    """
    statparams = ['max', 'min']

    # create multi-index: first is the selected channel, second is the stat param
    chans_selmax_ = []
    for k in chans_selmax: chans_selmax_.extend([k ,k])
    index = pd.MultiIndex.from_arrays([chans_selmax_, statparams*len(chans_selmax)])
    # initiale the table
    dfextr = pd.DataFrame(np.zeros((len(chans_selmax_), (len(chans_atmax)))),
                          index=index, columns=chans_atmax)

    # vectorised slightly faster for 8x8 channels
    # # 9.08 ms ± 72.9 µs per loop (mean ± std. dev. of 7 runs, 100 loops each)
    # select the index at which the max/minimum occurs for the channels in chans_selmax
    idxmax = dfres[chans_selmax].idxmax(axis=0).values
    idxmin = dfres[chans_selmax].idxmin(axis=0).values
    dfextr.loc[(chans_selmax, 'max'), chans_atmax] = dfres.loc[idxmax, chans_atmax].values
    dfextr.loc[(chans_selmax, 'min'), chans_atmax] = dfres.loc[idxmin, chans_atmax].values

    # # looping is slower for 8x8 channels
    # # 11.1 ms ± 21.1 µs per loop (mean ± std. dev. of 7 runs, 100 loops each)
    # for chan in chans_selmax:
    #     idxmax = dfres[chan].idxmax(axis=0)
    #     idxmin = dfres[chan].idxmin(axis=0)
    #     for atchan in chans_atmax:
    #         dfextr.loc[(chan, 'max'),atchan] = dfres.loc[idxmax, atchan]
    #         dfextr.loc[(chan, 'min'),atchan] = dfres.loc[idxmin, atchan]

    # dfextr.loc[pd.IndexSlice[:,'max'],chan] = dfres.iloc[imax, ichan]
    # dfextr.loc[pd.IndexSlice[:,'min'],chan] = dfres.iloc[imin, ichan]
    # dfextr.loc[pd.IndexSlice[:,:],:].values
    # assert np.allclose(dfextr.loc[:,:].values, dfextr2.loc[:,:].values)

    return dfextr


def envelope(res, int_env=False, Nx=300):

    # define which sensors to use by using the unique_ch_name feature
    ch_list = []

    # find all available outputs for tower, blades of type momentvec Mx
    sel = ((  (res.ch_df['bodyname']=='tower')
            | (res.ch_df['bodyname']=='blade1')
            | (res.ch_df['bodyname']=='blade2')
            | (res.ch_df['bodyname']=='blade3') )
           & (res.ch_df['sensortype']=='momentvec') & (res.ch_df['component']=='x')
          )
    nodeselx = res.ch_df[sel]['unique_ch_name'].values.tolist()

    # all momentvec My
    sel = ((  (res.ch_df['bodyname']=='tower')
            | (res.ch_df['bodyname']=='blade1')
            | (res.ch_df['bodyname']=='blade2')
            | (res.ch_df['bodyname']=='blade3') )
           & (res.ch_df['sensortype']=='momentvec') & (res.ch_df['component']=='y')
          )
    nodesely = set(res.ch_df[sel]['unique_ch_name'].values.tolist())

    for nodex in nodeselx:
        nodey = nodex.replace('-x', '-y')
        # make sure both Mx and My are defined
        if nodey not in nodesely:
            continue
        ch_list.append([nodex, nodey])

    fname = res.FileName + '_envelopes.h5'
    filt = tbl.Filters(complevel=9)
    with tbl.open_file(fname, mode="w", title=str(SIM_ID), filters=filt) as h5f:
        groupname = str(os.path.basename(res.FileName))
        groupname = groupname.replace('-', '_').replace('.', '_')
        ctab = h5f.create_group("/", groupname)

        # envelope = {}
        for ch_names in ch_list:

            ichans = []
            for ch_name in ch_names:
                ichans.append(res.ch_dict[ch_name]['chi'])
            cloud = res.sig[:, ichans]
            # Compute a Convex Hull, the vertices number varies according to
            # the shape of the poligon
            vertices = compute_envelope(cloud, int_env=int_env, Nx=Nx)
            # envelope[ch_names[0]] = vertices

            # # save as simple text file
            # title = ch_names[0]
            # fname = res.FileName.replace('res/', 'prepost-data/env_txt/') + f'_{title}.txt'
            # os.makedirs(os.path.dirname(fname), exist_ok=True)
            # np.savetxt(fname, vertices)

            # save as HDF5 table
            title = str(ch_names[0].replace('-', '_'))
            csv_table = h5f.create_table(ctab, title,
                                          EnvelopeClass.section,
                                          title=title)
            tablerow = csv_table.row
            for row in vertices: #envelope[ch_names[0]]:
                tablerow['Mx'] = float(row[0])
                tablerow['My'] = float(row[1])
                if len(row)>2:
                    tablerow['Mz'] = float(row[2])
                    if len(row)>3:
                        tablerow['Fx'] = float(row[3])
                        tablerow['Fy'] = float(row[4])
                        tablerow['Fz'] = float(row[5])
                    else:
                        tablerow['Fx'] = 0.0
                        tablerow['Fy'] = 0.0
                        tablerow['Fz'] = 0.0
                else:
                    tablerow['Mz'] = 0.0
                    tablerow['Fx'] = 0.0
                    tablerow['Fy'] = 0.0
                    tablerow['Fz'] = 0.0
                tablerow.append()
            csv_table.flush()
    # h5f.close()


def calc(fpath, do_envelope, no_bins, m, atmax):

    t0 = time.time()
    i0 = 0
    i1 = None
    ftype = '.csv'

    # remove the extension
    ext = fpath.split('.')[-1]
    if ext in ('sel', 'dat', 'hdf5'):
        fpath = '.'.join(fpath.split('.')[:-1])
    else:
        print('invalid file extension, ignored:', fpath)
        return

    fdir = os.path.dirname(fpath)
    fname = os.path.basename(fpath)
    try:
        res = windIO.LoadResults(fdir, fname, usecols=None, readdata=True)
    except Exception as e:
        print(f'loading {fpath} failed:')
        print(e)
        return
    neq = res.sig[-1,0] - res.sig[0,0]

    # add channels in memory so they can be included in the statistics
    # they are not saved back to disk
    res = add_channels(res)

    # convert to DataFrame
    dfres = res.sig2df()

    # save the envelope
    if do_envelope:
        envelope(res, int_env=False, Nx=300)

    # extremes and corresponding loads at the same time
    if atmax is not None:
        for node, val in atmax.items():
            dfextr = loads_at_extreme(dfres, val['chans_selmax'], val['chans_selmax'])
            dfextr.to_excel(fpath + f'_{node}_loads_at_extreme.xlsx')

    # statistics
    df_stats = res.statsdel_df(i0=i0, i1=i1, statchans='all', neq=neq,
                               no_bins=no_bins, m=m, delchans='all')

    # add path and fname as columns
    df_stats['case_id'] = fname
    df_stats['res_dir'] = fdir

    if ftype == '.csv':
        df_stats.to_csv(fpath+ftype)
    elif ftype == '.h5':
        df_stats.to_hdf(fpath+ftype, 'table', complib='zlib', complevel=9)
    print('% 7.03f  ' % (time.time() - t0), fname)


def par(logdir, resdir, cpus=30, chunksize=30, nostats=False, nolog=False,
        do_envelope=True, no_bins=46, m=[3,4,6,8,9,10,12], atmax=None):
    """Sophia has 32 CPUs per node.
    """

    # log file analysis
    if not nolog:
        files = glob(os.path.join(logdir, '**', '*.log'), recursive=True)
        print(f'start processing {len(files)} logfiles from dir: {logdir}')
        with Pool(processes=cpus) as pool:
            # chunksize = 10 #this may take some guessing ...
            for ind, res in enumerate(pool.imap(logcheck, files), chunksize):
                pass
    print()

    # consider both hdf5 and sel outputs
    if not nostats:
        files = glob(os.path.join(resdir, '**', '*.hdf5'), recursive=True)
        files.extend(glob(os.path.join(resdir, '**', '*.sel'), recursive=True))
        print(f'start processing {len(files)} resfiles from dir: {resdir}')
        # prepare the other arguments
        combis = [(k, do_envelope, no_bins, m, atmax) for k in files]

        # sequential loop
        # for combi in combis:
        #     calc(*combi)

        # start the parallal for-loop using X number of cpus
        with Pool(processes=cpus) as pool:
            # This method chops the iterable into a number of chunks which it submits
            # to the process pool as separate tasks. The (approximate) size of these
            # chunks can be specified by setting chunksize to a positive integer.
            for ind, res in enumerate(pool.starmap(calc, combis), chunksize):
                pass
    print()


def merge_stats(post_dir='prepost-data'):
    """Merge stats for each case into one stats table
    """

    print('start merging statistics into single table...')

    os.makedirs(post_dir, exist_ok=True)
    # -------------------------------------------------------------------------
    # MERGE POSTPRO ON NODE APPROACH INTO ONE DataFrame
    # -------------------------------------------------------------------------
    path_pattern = os.path.join('res', '**', '*.csv')
    csv_fname = '%s_statistics.csv' % SIM_ID
    # if zipchunks:
    #     path_pattern = os.path.join(post_dir, 'statsdel_chnk*.tar.xz')
    fcsv = os.path.join(post_dir, csv_fname)
    mdf = AppendDataFrames(tqdm=tqdm)
    cols = mdf.txt2txt(fcsv, path_pattern, tarmode='r:xz', header=0, sep=',',
                       header_fjoined=None, recursive=True)#, fname_col='[case_id]')
    # and convert to df: takes 2 minutes
    fdf = fcsv.replace('.csv', '.h5')
    store = pd.HDFStore(fdf, mode='w', complevel=9, complib='zlib')
    colnames = cols.split(',')
    # the first column is the channel name but the header is emtpy
    assert colnames[0] == ''
    colnames[0] = 'channel'
    dtypes = {col:np.float64 for col in colnames}
    dtypes['channel'] = str
    dtypes['case_id'] = str
    dtypes['res_dir'] = str
    # when using min_itemsize the column names should be valid variable names
    # mitemsize = {'channel':60, '[case_id]':60}
    mdf.csv2df_chunks(store, fcsv, chunksize=1000000, min_itemsize={}, sep=',',
                      colnames=colnames, dtypes=dtypes, header=0)
    store.close()
    print(f'\nsaved at: {fcsv}')


def merge_logs(post_dir='prepost-data'):

    print('start merging log analysis into single table...')

    os.makedirs(post_dir, exist_ok=True)
    # -------------------------------------------------------------------------
    # MERGE POSTPRO ON NODE APPROACH INTO ONE DataFrame
    # -------------------------------------------------------------------------
    lf = windIO.LogFile()
    path_pattern = os.path.join('logfiles', '**', '*.csv')
    # if zipchunks:
    #     path_pattern = os.path.join(POST_DIR, 'loganalysis_chnk*.tar.xz')
    csv_fname = '%s_ErrorLogs.csv' % SIM_ID
    fcsv = os.path.join(post_dir, csv_fname)
    mdf = AppendDataFrames(tqdm=tqdm)
    # individual log file analysis does not have header, make sure to include
    # a line for the header
    cols = mdf.txt2txt(fcsv, path_pattern, tarmode='r:xz', header=None,
                       header_fjoined=lf._header(), recursive=True)

    # FIXME: this is due to bug in log file analysis. What is going on here??
    # fix that some cases do not have enough columns
    with open(fcsv.replace('.csv', '2.csv'), 'w') as f1:
        with open(fcsv) as f2:
            for line in f2.readlines():
                if len(line.split(';'))==96:
                    line = line.replace(';0.00000000000;nan;-0.0000;',
                                        '0.00000000000;nan;-0.0000;')
                f1.write(line)

    # convert from CSV to DataFrame and save as Excel
    df = lf.csv2df(fcsv.replace('.csv', '2.csv'))
    # since this is mainly text it doesn't make any sense to convert to hdf5
    # for one DLB example, csv is 350 kB, hdf 2 MB.
    # df.to_hdf(fcsv.replace('.csv', '.h5'), 'table')
    df.to_excel(fcsv.replace('.csv', '.xlsx'))
    print(f"\nsaved at: {fcsv.replace('.csv', '2.csv')}")


def save_unique_chan_names(post_dir='prepost-data'):

    fdf = os.path.join(post_dir, '%s_statistics.h5' % SIM_ID)
    df_stats = pd.read_hdf(fdf, 'table')
    chans = df_stats['channel'].unique()
    # TODO: print table with HTC channel names on one side, and the unique
    # channel names on the other. Problem: one HTC line can have multiple channels
    # do not sort, leave in same order as in the sel file
    # chans.sort()
    fname = os.path.join(post_dir, '%s_unique-channel-names.csv' % SIM_ID)
    pd.DataFrame(chans).to_csv(fname)


if __name__ == '__main__':

    parser = ArgumentParser(description = "Parallel post-processing with the "
                            "Python build-in multiprocessing module.")
    parser.add_argument('-n', type=int, default=2, action='store',
                        dest='cpus', help='Number of CPUs to use on the node')
    parser.add_argument('-c', type=int, default=80, action='store',
                        dest='chunksize', help='Chunksize Pool.')

    parser.add_argument('--res', type=str, default='res', action='store',
                        dest='resdir', help='Directory containing result files')
    parser.add_argument('--log', type=str, default='log', action='store',
                        dest='logdir', help='Directory containing HAWC2 log files')

    parser.add_argument('--nologs', default=False, action='store_true',
                        dest='nologs', help='Do not perform the logfile analysis.')
    parser.add_argument('--nostats', default=False, action='store_true',
                        dest='nostats', help='Do not calculate statistics.')
    parser.add_argument('--envelope', default=False, action='store_true',
                        dest='do_envelope', help='Calculate envelopes.')
    parser.add_argument('--no_bins', default=46, action='store',
                        dest='no_bins', help='Number of bins for the rainflow counting.')
    parser.add_argument('--atmax', default=False, action='store', type=str,
                        dest='atmax', help='File name that holds the channels to create '
                        'the table at which maxima and corresponding values are selected.')
    opt = parser.parse_args()

    if opt.atmax:
        with open(opt.atmax) as fio:
            atmax_js = json.load(fio)

    SIM_ID = os.path.basename(os.getcwd())
    par(opt.logdir, opt.resdir, cpus=opt.cpus, chunksize=opt.chunksize,
        nolog=opt.nologs, nostats=opt.nostats, do_envelope=opt.do_envelope,
        no_bins=opt.no_bins, m=[3,4,6,8,9,10,12], atmax=atmax_js)
    if not opt.nostats:
        merge_stats(post_dir='prepost-data')
        save_unique_chan_names(post_dir='prepost-data')
    if not opt.nologs:
        merge_logs(post_dir='prepost-data')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun 23 11:07:50 2021

@author: dave
"""

# from os.path import join as pjoin

import numpy as np
from lxml import objectify#, etree)

# from matplotlib import pyplot as plt
# import pandas as pd

# import wetb
from wetb.prepost import misc
# from wetb.hawc2 import (HTCFile, AEFile, PCFile, StFile)


# TODO: this dictionary utility could go to a dict sub-class in misc?
def find_next_unique_key(d, key):
    k = 1
    while True:
        k += 1
        if key + f'__{k}' not in d:
            key = key + f'__{k}'
            return key


# TODO: this dictionary utility could go to a dict sub-class in misc?
def add_new_unique_key(key, values, d):
    """
    Add key:values to dictionary d. If key already occurs, append __x
    where x is the number of previous occurances of key(__x).

    Parameters
    ----------
    key : str
        DESCRIPTION.
    values : list
        DESCRIPTION.
    d : dict
        Dictionary.

    Returns
    -------
    key : str
        DESCRIPTION.
    d : dict
        DESCRIPTION.

    """
    if key in d:
        key = find_next_unique_key(d, key)
    d[key] = [values]
    return key, d


class ReadBladedProject:

    def __init__(self, fname):

        with open(fname, encoding='utf-8') as fobj:
            xml_str = fobj.read().encode('utf-8')
        self.bd, self.xmlroot = self.parse_bladeddata(xml_str)

        self.set_attr_and_check()

        # some things are just a little different
        # TMASS has a list of materials and their properties
        # get rid of the quotes
        # tmp = [el.replace("'", '') for el in self.bd['TMASS']['MATERIAL'][0]]
        # self.bd['TMASS']['MATERIAL'] = tmp
        unique_mat = set(self.get_key('TMASS', 'MATERIAL').flatten().tolist())
        self.tow_mat_prop = {k:self.get_key('TMASS', k) for k in unique_mat}
        # material_props = {}
        # for k in unique_mat:
        #     material_props[k.replace("'", '')] = self.get_key('TMASS', k)

    def parse_bladeddata(self, xml_str):
        """
        The XML field BladedData contains what seems like the main core input
        data for BLADED, and formatted in some structured way.

        Parameters
        ----------
        xml_str : TYPE
            DESCRIPTION.

        Returns
        -------
        bd : TYPE
            DESCRIPTION.
        xmlroot : TYPE
            DESCRIPTION.

        """

        # root = etree.fromstring(xml)
        # elems = root.getchildren()
        # bladeddata = elems[1].text

        xmlroot = objectify.fromstring(xml_str)

        # the non-xml formatted BLADED model is situated in a non-xml field
        # called BladedData
        bd = {}
        mstart = None
        for i, line in enumerate(xmlroot.BladedData.text.split('\n')):

            # TODO: values embedded in double quotes (") can contain entire
            # MSTART/MEND sub-sections (so embedded sections)

            # split replace tabs with spaces, split on spaces, remove empty
            linels = misc.remove_items(line.replace('\t', ' ').split(' '), '')
            # commas can also be separators, in addition to spaces
            linels2 = []
            for k in linels:
                linels2.extend(k.split(','))
            linels = misc.remove_items(linels2, '')

            # ignore empty lines
            if len(linels) < 1:
                continue

            # entries can be numbers if the previous key contains multiple data points
            try:
                float(linels[0])
                el0isnum = True
            except ValueError:
                el0isnum = False

            # start of a sub-section that contains (non-unique) keys as well
            if linels[0].upper().startswith('MSTART'):
                mtag = linels[-1]
                mstart = {}

            # at the end of the sub-section add the sub-section to the main dict
            elif linels[0].upper().startswith('MEND'):
                # FIXME: implement MSTART sections embedded in double quoted values
                try:
                    # if the section key is not unique, make it so by appending __X
                    if mtag in bd:
                        mtag = find_next_unique_key(bd, mtag)
                    bd[mtag] = mstart
                except UnboundLocalError:
                    print('warning: ignored embedded mstart/mend section')
                    print(f'at line: {i+1}')
                mstart = None

            # if we are under a sub-section
            elif mstart is not None:
                # if the line contains a keyword
                if not el0isnum:
                    tag, mstart = add_new_unique_key(linels[0], linels[1:], mstart)
                # line is datapoint that needs to be added to key that occured before
                else:
                    mstart[tag].append(linels)

            # add numerical values to key that occured before
            elif el0isnum:
                 bd[tag].append(linels)

            else:
                tag, bd = add_new_unique_key(linels[0], linels[1:], bd)

        return bd, xmlroot

    def get_key(self, key1, key2=False):
        """
        Get key from the BladedData CDATA section and format to int32 or
        float32 numpy arrays if possible withouth precision loss.

        Parameters
        ----------
        key1 : str
            DESCRIPTION.
        key2 : str, optional
            DESCRIPTION. The default is False.


        Returns
        -------
        numpy.array
            Values from key1/key2 formatted as a numpy array. Converted to
            numpy.int32, numpy.float32 if possible withouth precision loss,
            otherwise an object array is returned.

        """

        if key1 not in self.bd:
            raise KeyError(f'{key1} not found in BLADED file.')

        if key2 is not False:
            if key2 not in self.bd[key1]:
                raise KeyError(f'{key2} not found in MSTART {key1} of BLADED file.')
            data = self.bd[key1][key2]
        else:
            data = self.bd[key1]

        # in case we defined a mstart key
        if isinstance(data, dict):
            return data

        # i ,j = len(data), len(data[0])

        # FIXME: this is a very expensive way of converting it, but it might
        # not matter at all since very little model data is actually considered
        data_arr = np.array(data)
        try:
            data_arr = data_arr.astype(np.int32)
        except ValueError:
            try:
                data_arr = data_arr.astype(np.float32)
            except ValueError:
                pass

        return data_arr

        # return np.array(data, dtype=np.float32)

        # if isinstance(data[0], list) and len(data[0]) == 1:
        #     data = float(data[0])
        # if isinstance(data[0], list) and len(data[0]) > 1:
        #     data_arr = np.array(data, dtype=np.float32)

    def set_attr_and_check(self):
        """Check that BGEOMMB, BMASSMB, BSTIFFMB has indeed the same node
        repated twice every time.
        """

        # self.bd['BGEOMMB'].keys(), but only those relevant
        keysg = ['RJ', 'DIST', 'REF_X', 'REF_Y', 'CHORD', 'TWIST', 'CE_X',
                 'CE_Y', 'BTHICK', 'FOIL', 'MOVING']
        nbe = self.get_key('BGEOMMB', 'NBE')[0,0]

        # self.bd['BMASSMB'].keys(), but only those relevant
        keysm = ['CM_X', 'CM_Y', 'MASS', 'SINER', 'RGRATIO', 'BETA_M']

        # self.bd['BSTIFFMB'].keys(), but only those relevant
        keyss = ['EIFLAP', 'EIEDGE', 'BETA_S', 'GJ', 'CS_X', 'CS_Y', 'GAFLAP',
                 'GAEDGE']

        mkeys = ['BGEOMMB', 'BMASSMB', 'BSTIFFMB']
        for mkey, keys in zip(mkeys, [keysg, keysm, keyss]):
            for key in keys:
                res = self.get_key(mkey, key)
                try:
                    assert np.allclose(res[0,0::2], res[0,1::2])
                except TypeError:
                    # allclose doesn't make sense for text arrays
                    assert np.compare_chararrays(res[0,0::2], res[0,1::2],
                                                 '==', True).all()
                assert res.shape[1]==nbe
                if hasattr(self, key.lower()):
                    raise UserWarning(key, 'already exists')
                setattr(self, key.lower(), res[0,0::2])

    def print_xml_tree(self, fname):
        """For discovery purposes: print full tree + values/text


        Parameters
        ----------
        fname : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """

        # def print_rec(root):
        #     # if hasattr(root, 'getparent'):
        #     #     print(root.getparent().tag.title, end='.')
        #     #     print_rec(root.getparent())
        #     for el in root.getchildren():
        #         print(print_rec(el))

        # def print_root_tree(root):
        #     root.getparent()
        #     print()

        # tree = etree.fromstring(xml_str)
        # els = tree.xpath('/')
        # for el in els:
        #     print(el)

        # tree = etree.fromstring(xml_str)

        # xmlroot = objectify.fromstring(xml_str)
        # with open(fname+'.structure', 'w') as f:
        #     for line in xmlroot.descendantpaths():
        #         f.write(line + '\n')

        # Recursive XML parsing python using ElementTree
        # https://stackoverflow.com/q/28194703/3156685

        roottree = self.xmlroot.getroottree()
        def print_tree_recursive(root):
            # print(roottree.getpath(root), end=' : ')
            # print(root.tag.title())
            f.write(roottree.getpath(root) + ' : ')
            f.writelines(root.values())
            if root.text is not None:
                f.write(' : ' + root.text)
            f.write('\n')
            for elem in root.getchildren():
                print_tree_recursive(elem)
        with open(fname+'.structure.value', 'w') as f:
            for el in self.xmlroot.getchildren():
                if el.tag.title()!='Bladeddata':
                    print_tree_recursive(el)

    # def get_frequencies(self):
    #     """
    #     """

    #     blades = self.xmlroot.NewGUI.Turbine.Blades.FlexibilityModel

    #     blades.Settings.ModesWithDampingDefined
    #     blades.PartModes # with lots of BladeModeContainer
    #     blades.WholeBladeModes
    #     blades.WholeBladeModes.WholeBladeMode

    #     blades.WholeBladeModes.WholeBladeMode.Name
    #     blades.WholeBladeModes.WholeBladeMode.Frequency
    #     blades.WholeBladeModes.WholeBladeMode.Damping
    #     blades.WholeBladeModes.WholeBladeMode.Components
    #     len(blades.WholeBladeModes.WholeBladeMode.Components.getchildren())
    #     # 60 elements

    #     for wholeblademode in blades.WholeBladeModes.iterchildren():
    #         print(wholeblademode.Name)
    #         print(wholeblademode.Frequency, wholeblademode.Damping)

    #     tower = self.xmlroot.NewGUI.Turbine.SupportStructure.FlexibilityModel
    #     for towermode in tower.Modes.getchildren():
    #         print(towermode.Description)
    #         towermode.Frequency
    #         towermode.Damping
